import json
import os

import numpy as np

from .. import resolvers, util
from ..constants import log
from .dae import _collada_exporters
from .gltf import export_glb, export_gltf
from .obj import export_obj
from .off import _off_exporters
from .ply import _ply_exporters
from .stl import export_stl, export_stl_ascii
from .threemf import _3mf_exporters
from .urdf import export_urdf  # NOQA
from .xyz import _xyz_exporters


def export_mesh(mesh, file_obj, file_type=None, resolver=None, **kwargs):
    """
    Export a Trimesh object to a file- like object, or to a filename

    Parameters
    -----------
    file_obj : str, file-like
      Where should mesh be exported to
    file_type : str or None
      Represents file type (eg: 'stl')
    resolver : None or trimesh.resolvers.Resolver
      Resolver to write referenced assets to

    Returns
    ----------
    exported : bytes or str
      Result of exporter
    """
    # if we opened a file object in this function
    # we will want to close it when we're done
    was_opened = False
    file_name = None

    if util.is_pathlib(file_obj):
        # handle `pathlib` objects by converting to string
        file_obj = str(file_obj.absolute())

    if util.is_string(file_obj):
        if file_type is None:
            # get file type from file name
            file_type = (str(file_obj).split(".")[-1]).lower()
        if file_type in _mesh_exporters:
            was_opened = True
            file_name = file_obj
            # get full path of file before opening
            file_path = os.path.abspath(os.path.expanduser(file_obj))
            file_obj = open(file_path, "wb")
            if resolver is None:
                # create a resolver which can write files to the path
                resolver = resolvers.FilePathResolver(file_path)

    # make sure file type is lower case
    file_type = str(file_type).lower()

    if file_type not in _mesh_exporters:
        raise ValueError("%s exporter not available!", file_type)

    if isinstance(mesh, (list, tuple, set, np.ndarray)):
        faces = 0
        for m in mesh:
            faces += len(m.faces)
        log.debug(
            "Exporting %d meshes with a total of %d faces as %s",
            len(mesh),
            faces,
            file_type.upper(),
        )
    elif hasattr(mesh, "faces"):
        # if the mesh has faces log the number
        log.debug("Exporting %d faces as %s", len(mesh.faces), file_type.upper())

    # OBJ files save assets everywhere
    if file_type == "obj":
        kwargs["resolver"] = resolver

    # run the exporter
    export = _mesh_exporters[file_type](mesh, **kwargs)

    # if the export is multiple files (i.e. GLTF)
    if isinstance(export, dict):
        # if we have a filename rename the default GLTF
        if file_name is not None and "model.gltf" in export:
            export[os.path.basename(file_name)] = export.pop("model.gltf")

        # write the files if a resolver has been passed
        if resolver is not None:
            for name, data in export.items():
                resolver.write(name=name, data=data)

        return export

    if hasattr(file_obj, "write"):
        result = util.write_encoded(file_obj, export)
    else:
        result = export

    # if we opened anything close it here
    if was_opened:
        file_obj.close()

    return result


def export_dict64(mesh):
    """
    Export a mesh as a dictionary, with data encoded
    to base64.
    """
    return export_dict(mesh, encoding="base64")


def export_dict(mesh, encoding=None):
    """
    Export a mesh to a dict

    Parameters
    ------------
    mesh : trimesh.Trimesh
      Mesh to be exported
    encoding : str or None
      Such as 'base64'

    Returns
    -------------
    export : dict
      Data stored in dict
    """

    def encode(item, dtype=None):
        if encoding is None:
            return item.tolist()
        else:
            if dtype is None:
                dtype = item.dtype
            return util.array_to_encoded(item, dtype=dtype, encoding=encoding)

    # metadata keys we explicitly want to preserve
    # sometimes there are giant datastructures we don't
    # care about in metadata which causes exports to be
    # extremely slow, so skip all but known good keys
    meta_keys = ["units", "file_name", "file_path"]
    metadata = {k: v for k, v in mesh.metadata.items() if k in meta_keys}

    export = {
        "metadata": metadata,
        "faces": encode(mesh.faces),
        "face_normals": encode(mesh.face_normals),
        "vertices": encode(mesh.vertices),
    }
    if mesh.visual.kind == "face":
        export["face_colors"] = encode(mesh.visual.face_colors)
    elif mesh.visual.kind == "vertex":
        export["vertex_colors"] = encode(mesh.visual.vertex_colors)

    return export


def scene_to_dict(scene, use_base64=False, include_metadata=True):
    """
    Export a Scene object as a dict.

    Parameters
    -------------
    scene : trimesh.Scene
      Scene object to be exported

    Returns
    -------------
    as_dict : dict
      Scene as a dict
    """

    # save some basic data about the scene
    export = {
        "graph": scene.graph.to_edgelist(),
        "geometry": {},
        "scene_cache": {
            "bounds": scene.bounds.tolist(),
            "extents": scene.extents.tolist(),
            "centroid": scene.centroid.tolist(),
            "scale": scene.scale,
        },
    }

    if include_metadata:
        try:
            # jsonify will convert numpy arrays to lists recursively
            # a little silly round-tripping to json but it is pretty fast
            export["metadata"] = json.loads(util.jsonify(scene.metadata))
        except BaseException:
            log.warning("failed to serialize metadata", exc_info=True)

    # encode arrays with base64 or not
    if use_base64:
        file_type = "dict64"
    else:
        file_type = "dict"

    # if the mesh has an export method use it
    # otherwise put the mesh itself into the export object
    for geometry_name, geometry in scene.geometry.items():
        if hasattr(geometry, "export"):
            # export the data
            exported = {
                "data": geometry.export(file_type=file_type),
                "file_type": file_type,
            }
            export["geometry"][geometry_name] = exported
        else:
            # case where mesh object doesn't have exporter
            # might be that someone replaced the mesh with a URL
            export["geometry"][geometry_name] = geometry
    return export


def export_scene(scene, file_obj, file_type=None, resolver=None, **kwargs):
    """
    Export a snapshot of the current scene.

    Parameters
    ----------
    file_obj : str, file-like, or None
      File object to export to
    file_type : str or None
      What encoding to use for meshes
      IE: dict, dict64, stl

    Returns
    ----------
    export : bytes
      Only returned if file_obj is None
    """
    if len(scene.geometry) == 0:
        raise ValueError("Can't export empty scenes!")

    if util.is_pathlib(file_obj):
        # handle `pathlib` objects by converting to string
        file_obj = str(file_obj.absolute())

    # if we weren't passed a file type extract from file_obj
    if file_type is None:
        if util.is_string(file_obj):
            file_type = str(file_obj).split(".")[-1]
        else:
            raise ValueError("file_type not specified!")

    # always remove whitespace and leading characters
    file_type = file_type.strip().lower().lstrip(".")

    # now handle our different scene export types
    if file_type == "gltf":
        data = export_gltf(scene, **kwargs)
    elif file_type == "glb":
        data = export_glb(scene, **kwargs)
    elif file_type == "dict":
        data = scene_to_dict(scene, *kwargs)
    elif file_type == "obj":
        # if we are exporting by name automatically create a
        # resolver which lets the exporter write assets like
        # the materials and textures next to the exported mesh
        if resolver is None and util.is_string(file_obj):
            resolver = resolvers.FilePathResolver(file_obj)
        data = export_obj(scene, resolver=resolver, **kwargs)
    elif file_type == "dict64":
        data = scene_to_dict(scene, use_base64=True)
    elif file_type == "svg":
        from trimesh.path.exchange import svg_io

        data = svg_io.export_svg(scene, **kwargs)
    elif file_type == "ply":
        data = _mesh_exporters["ply"](scene.dump(concatenate=True), **kwargs)
    elif file_type == "stl":
        data = export_stl(scene.dump(concatenate=True), **kwargs)
    elif file_type == "3mf":
        data = _mesh_exporters["3mf"](scene, **kwargs)
    else:
        raise ValueError(f"unsupported export format: {file_type}")

    # now write the data or return bytes of result
    if isinstance(data, dict):
        # GLTF files return a dict-of-bytes as they
        # represent multiple files so create a filepath
        # resolver and write the files if someone passed
        # a path we can write to.
        if resolver is None and util.is_string(file_obj):
            resolver = resolvers.FilePathResolver(file_obj)
            # the requested "gltf"
            bare_path = os.path.split(file_obj)[-1]
            for name, blob in data.items():
                if name == "model.gltf":
                    # write the root data to specified file
                    resolver.write(bare_path, blob)
                else:
                    # write the supporting files
                    resolver.write(name, blob)
        return data

    if hasattr(file_obj, "write"):
        # if it's just a regular file object
        return util.write_encoded(file_obj, data)
    elif util.is_string(file_obj):
        # assume strings are file paths
        file_path = os.path.abspath(os.path.expanduser(file_obj))
        with open(file_path, "wb") as f:
            util.write_encoded(f, data)

    # no writeable file object so return data
    return data


_mesh_exporters = {
    "stl": export_stl,
    "dict": export_dict,
    "glb": export_glb,
    "obj": export_obj,
    "gltf": export_gltf,
    "dict64": export_dict64,
    "stl_ascii": export_stl_ascii,
}
_mesh_exporters.update(_ply_exporters)
_mesh_exporters.update(_off_exporters)
_mesh_exporters.update(_collada_exporters)
_mesh_exporters.update(_xyz_exporters)
_mesh_exporters.update(_3mf_exporters)



try:
    from . import generic as g
except BaseException:
    import generic as g


class RegistrationTest(g.unittest.TestCase):
    def test_procrustes(self):
        # every combination of possible boolean options
        # a_flip and a_scale are apply-to-test-data
        opt = list(g.itertools.combinations([True, False] * 6, 6))
        # generate deterministic but random-ish transforms
        matrices = iter(g.random_transforms(count=len(opt)))

        for reflection, translation, scale, a_flip, a_scale, weight in opt:
            # create random points in space
            points_a = (g.random((1000, 3)) - 0.5) * 1000
            # get a random transform from the iterator
            matrix = next(matrices)
            # apply a flip (reflection) to test data
            if a_flip:
                matrix = g.np.dot(
                    matrix,
                    g.trimesh.transformations.reflection_matrix(
                        point=[0, 0, 0], normal=[0, 1, 0]
                    ),
                )
            # apply scale to test data
            if a_scale:
                matrix = g.np.dot(
                    matrix, g.trimesh.transformations.scale_matrix(0.1235234)
                )
            # apply transform to points A
            points_b = g.trimesh.transform_points(points_a, matrix)

            # weight points or not
            if weight:
                weights = (g.random(len(points_a)) + 9) / 10
            else:
                weights = None

            # run the solver
            (matrixN, transformed, cost) = g.trimesh.registration.procrustes(
                points_a,
                points_b,
                reflection=reflection,
                translation=translation,
                scale=scale,
                weights=weights,
            )
            # if we're not weighting the results
            # should be identical with None vs all-ones
            (matrixN_C, transformed_C, cost_C) = g.trimesh.registration.procrustes(
                points_a,
                points_b,
                reflection=reflection,
                translation=translation,
                scale=scale,
                weights=g.np.ones(len(points_a)),
            )
            if weight:
                # weights should have changed the matrix
                assert not g.np.allclose(matrixN, matrixN_C)
            else:
                # no weights so everything should be identical
                assert g.np.allclose(matrixN, matrixN_C)
                assert g.np.allclose(transformed_C, transformed)
                assert g.np.isclose(cost, cost_C)

            # the points should be identical if the function
            # was allowed to translate in space
            # and there were no weights, scaling, or reflection
            identical = (
                translation
                and (not weight)
                and (not a_flip or reflection)
                and (not a_scale or scale)
            )

            if identical:
                assert cost < 0.001
                # it should have found the matrix we used
                assert g.np.allclose(matrixN, matrix)

            # if reflection is not allowed, the determinant
            # should always be close to 1.0 for the rotation
            det = g.np.linalg.det(matrixN[:3, :3])
            # reflection not allowed
            if not reflection:
                # determinant should always be positive
                assert det > 1e-8
                if not scale:
                    # no reflection or scale means determinant
                    # of 1.0
                    assert g.np.isclose(det, 1.0)

            # scaling not allowed
            if not scale:
                # allowed to be -1.0 or 1.0
                assert g.np.isclose(g.np.abs(det), 1.0)

            # points have a flip applied
            # procrustes is allowed to use reflection
            # and there is no scaling in the matrix
            if a_flip and reflection and not scale:
                assert g.np.isclose(det, -1.0)

    def test_icp_mesh(self):
        # see if ICP alignment works with meshes
        m = g.trimesh.creation.box()
        X = m.sample(10)
        X = X + [0.1, 0.1, 0.1]
        _matrix, _transformed, cost = g.trimesh.registration.icp(X, m, scale=False)
        assert cost < 0.01

    def test_icp_points(self):
        # see if ICP alignment works with point clouds
        # create random points in space
        points_a = (g.random((1000, 3)) - 0.5) * 1000
        # create a random transform
        # matrix = g.trimesh.transformations.random_rotation_matrix()
        # create a small transform
        # ICP will not work at all with large transforms
        matrix = g.trimesh.transformations.rotation_matrix(g.np.radians(1.0), [0, 0, 1])

        # take a few randomly chosen points and make
        # sure the order is permutated
        index = g.np.random.choice(g.np.arange(len(points_a)), 20)
        # transform and apply index
        points_b = g.trimesh.transform_points(points_a[index], matrix)
        # tun the solver
        matrixN, transformed, cost = g.trimesh.registration.icp(points_b, points_a)
        assert cost < 1e-3
        assert g.np.allclose(matrix, g.np.linalg.inv(matrixN))
        assert g.np.allclose(transformed, points_a[index])

    def test_mesh(self):
        noise = 0.05
        extents = [6, 12, 3]

        # create the mesh as a simple box
        mesh = g.trimesh.creation.box(extents=extents)

        # subdivide until we have more faces than we want
        for _i in range(3):
            mesh = mesh.subdivide()
        # apply tessellation and random noise
        mesh = mesh.permutate.noise(noise)
        # randomly rotation with translation
        transform = g.trimesh.transformations.random_rotation_matrix()
        transform[:3, 3] = (g.random(3) - 0.5) * 1000

        mesh.apply_transform(transform)

        scan = mesh
        # create a "true" mesh
        truth = g.trimesh.creation.box(extents=extents)

        for a, b in [[truth, scan], [scan, truth]]:
            a_to_b, _cost = a.register(b)

            a_check = a.copy()
            a_check.apply_transform(a_to_b)

            assert g.np.linalg.norm(a_check.centroid - b.centroid) < (noise * 2)

            # find the distance from the truth mesh to each scan vertex
            distance = a_check.nearest.on_surface(b.vertices)[1]
            assert distance.max() < (noise * 2)

        # try our registration with points
        points = g.trimesh.transform_points(
            scan.sample(100), matrix=g.trimesh.transformations.random_rotation_matrix()
        )
        truth_to_points, _cost = truth.register(points)
        truth.apply_transform(truth_to_points)
        distance = truth.nearest.on_surface(points)[1]

        assert distance.mean() < noise

    def test_nricp(self):
        # Get two meshes that have a comparable shape
        source = g.get_mesh("reference.obj", process=False)
        target = g.get_mesh("target.obj", process=False)

        # Vertex indices of landmarks source / target
        landmarks_vertex_indices = g.np.array(
            [
                [177, 1633],
                [181, 1561],
                [614, 1556],
                [610, 1629],
                [114, 315],
                [398, 413],
                [812, 412],
                [227, 99],
                [241, 87],
                [674, 86],
                [660, 98],
                [362, 574],
                [779, 573],
            ]
        )

        source_markers_vertices = source.vertices[landmarks_vertex_indices[:, 0]]
        target_markers_vertices = target.vertices[landmarks_vertex_indices[:, 1]]

        T = g.trimesh.registration.procrustes(
            source_markers_vertices, target_markers_vertices
        )[0]
        source.vertices = g.trimesh.transformations.transform_points(source.vertices, T)

        # Just for the sake of using barycentric coordinates...
        use_barycentric_coordinates = True
        if use_barycentric_coordinates:
            source_markers_vertices = source.vertices[landmarks_vertex_indices[:, 0]]
            source_markers_tids = g.trimesh.proximity.closest_point(
                source, source_markers_vertices
            )[2]
            source_markers_barys = g.trimesh.triangles.points_to_barycentric(
                source.triangles[source_markers_tids], source_markers_vertices
            )
            source_landmarks = (source_markers_tids, source_markers_barys)
        else:
            source_landmarks = landmarks_vertex_indices[:, 0]
        # Make copies to check nothing is modified
        source_copy = source.copy()
        target_copy = target.copy()

        # Parameters for nricp_amberg
        wl = 3
        max_iter = 10
        wn = 0.5
        steps_amberg = [
            # ws, wl, wn, max_iter
            [0.02, wl, wn, max_iter],
            [0.007, wl, wn, max_iter],
            [0.002, wl, wn, max_iter],
        ]

        # Parameters for nricp_sumner
        wl = 1000
        wi = 0.0001
        ws = 1
        wn = 1
        steps_sumner = [
            # wc, wi, ws, wl, wn
            [0.1, wi, ws, wl, wn],
            [0.1, wi, ws, wl, wn],
            [1, wi, ws, wl, wn],
            [5, wi, ws, wl, wn],
            [10, wi, ws, wl, wn],
            [50, wi, ws, wl, wn],
            [100, wi, ws, wl, wn],
            [1000, wi, ws, wl, wn],
            [10000, wi, ws, wl, wn],
        ]
        # Amberg et. al 2007
        records_amberg_no_ldm = g.trimesh.registration.nricp_amberg(
            source,
            target,
            distance_threshold=0.05,
            steps=steps_amberg,
            return_records=True,
        )
        records_amberg_ldm = g.trimesh.registration.nricp_amberg(
            source,
            target,
            source_landmarks=source_landmarks,
            target_positions=target_markers_vertices,
            steps=steps_amberg,
            return_records=True,
            distance_threshold=0.05,
        )
        try:
            g.trimesh.registration.nricp_amberg(source, target)
        except KeyError:
            raise AssertionError()  # related to #1724

        # Sumner and Popovic 2004
        records_sumner_no_ldm = g.trimesh.registration.nricp_sumner(
            source,
            target,
            distance_threshold=0.05,
            steps=steps_sumner,
            return_records=True,
        )
        records_sumner_ldm = g.trimesh.registration.nricp_sumner(
            source,
            target,
            source_landmarks=source_landmarks,
            target_positions=target_markers_vertices,
            steps=steps_sumner,
            return_records=True,
            distance_threshold=0.05,
        )
        try:
            g.trimesh.registration.nricp_sumner(source, target)
        except KeyError:
            raise AssertionError()  # related to #1724

        d_amberg_no_ldm = g.trimesh.proximity.closest_point(
            target, records_amberg_no_ldm[-1]
        )[1]
        d_amberg_ldm = g.trimesh.proximity.closest_point(target, records_amberg_ldm[-1])[
            1
        ]
        d_sumner_no_ldm = g.trimesh.proximity.closest_point(
            target, records_sumner_no_ldm[-1]
        )[1]
        d_sumner_ldm = g.trimesh.proximity.closest_point(target, records_sumner_ldm[-1])[
            1
        ]

        # Meshes should remain untouched
        assert g.np.allclose(source.vertices, source_copy.vertices)
        assert g.np.allclose(target.vertices, target_copy.vertices)

        # Basically a lot of check to check if the return values change somehow
        assert d_amberg_no_ldm.min() < 1e-7
        assert d_amberg_no_ldm.max() > 0.05
        assert d_amberg_no_ldm.mean() < 1e-3

        assert d_amberg_ldm.min() > 1e-7
        assert d_amberg_ldm.max() < 0.06
        assert d_amberg_ldm.mean() < 1e-3

        assert d_sumner_no_ldm.min() < 1e-8
        assert d_sumner_no_ldm.max() < 0.03
        assert d_sumner_no_ldm.mean() < 1e-3

        assert d_sumner_ldm.min() < 1e-8
        assert d_sumner_ldm.max() > 0.05
        assert d_sumner_ldm.mean() < 1e-3

        dl_amberg_no_ldm = g.np.linalg.norm(
            records_amberg_no_ldm[-1][landmarks_vertex_indices[:, 0]]
            - target_markers_vertices,
            axis=-1,
        )
        dl_amberg_ldm = g.np.linalg.norm(
            records_amberg_ldm[-1][landmarks_vertex_indices[:, 0]]
            - target_markers_vertices,
            axis=-1,
        )
        dl_sumner_no_ldm = g.np.linalg.norm(
            records_sumner_no_ldm[-1][landmarks_vertex_indices[:, 0]]
            - target_markers_vertices,
            axis=-1,
        )
        dl_sumner_ldm = g.np.linalg.norm(
            records_sumner_ldm[-1][landmarks_vertex_indices[:, 0]]
            - target_markers_vertices,
            axis=-1,
        )

        assert dl_amberg_no_ldm.min() > 0.01
        assert dl_amberg_no_ldm.max() > 0.1
        assert dl_amberg_no_ldm.mean() > 0.05

        assert dl_amberg_ldm.min() < 0.0002
        assert dl_amberg_ldm.max() < 0.001
        assert dl_amberg_ldm.mean() < 0.0004

        assert dl_sumner_no_ldm.min() > 0.01
        assert dl_sumner_no_ldm.max() > 0.1
        assert dl_sumner_no_ldm.mean() > 0.05

        assert dl_sumner_ldm.min() < 0.0007
        assert dl_sumner_ldm.max() < 0.006
        assert dl_sumner_ldm.mean() < 0.004

    def test_query_from_points(self):
        points = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        query_point = [[0, 0.5, 0]]
        qres = g.trimesh.registration._from_points(
            points, query_point, return_normals=False
        )
        assert qres["vertex_indices"][0] == 1
        assert g.np.all(qres["nearest"][0] == [0, 1, 0])
        assert "normals" not in qres
        qres = g.trimesh.registration._from_points(
            points, query_point, return_normals=True
        )
        normal = g.np.ones(3)
        normal = normal / g.np.linalg.norm(normal)
        assert g.np.allclose(qres["normals"][0], normal) or g.np.allclose(
            qres["normals"][0], -normal
        )

    def test_query_from_mesh(self):
        points = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        faces = [[0, 1, 2]]
        mesh = g.trimesh.Trimesh(vertices=points, faces=faces)
        query_point = [[0, 0.5, 0]]
        qres = g.trimesh.registration._from_mesh(
            mesh,
            query_point,
            return_barycentric_coordinates=False,
            return_normals=False,
            return_interpolated_normals=False,
        )

        assert "normals" not in qres
        assert "barycentric_coordinates" not in qres
        assert "interpolated_normals" not in qres

        qres = g.trimesh.registration._from_mesh(
            mesh,
            query_point,
            return_barycentric_coordinates=False,
            return_normals=False,
            return_interpolated_normals=True,
        )

        assert "normals" not in qres
        assert "barycentric_coordinates" in qres
        assert "interpolated_normals" in qres


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
threedxml.py
-------------

Load 3DXML files, a scene format from Dassault products like Solidworks, Abaqus, Catia
"""

import numpy as np

try:
    # `pip install pillow`
    # optional: used for textured meshes
    from PIL import Image
except BaseException as E:
    # if someone tries to use Image re-raise
    # the import error so they can debug easily
    from ..exceptions import ExceptionWrapper

    Image = ExceptionWrapper(E)

import collections
import json

from .. import util
from ..visual.texture import TextureVisuals


def load_3DXML(file_obj, *args, **kwargs):
    """
    Load a 3DXML scene into kwargs. 3DXML is a CAD format
    that can be exported from Solidworks

    Parameters
    ------------
    file_obj : file object
      Open and containing 3DXML data

    Returns
    -----------
    kwargs : dict
      Can be passed to trimesh.exchange.load.load_kwargs
    """
    archive = util.decompress(file_obj, file_type="zip")

    # a dictionary of file name : lxml etree
    as_etree = {}
    for k, v in archive.items():
        # wrap in try statement, as sometimes 3DXML
        # contains non- xml files, like JPG previews
        try:
            as_etree[k] = etree.XML(v.read())
        except etree.XMLSyntaxError:
            # move the file object back to the file start
            v.seek(0)

    # the file name of the root scene
    root_file = as_etree["Manifest.xml"].find("{*}Root").text
    # the etree of the scene layout
    tree = as_etree[root_file]
    # index of root element of directed acyclic graph
    root_id = tree.find("{*}ProductStructure").attrib["root"]

    # load the materials library from the materials elements
    colors = {}
    images = {}
    # but only if it exists
    material_key = "CATMaterialRef.3dxml"
    if material_key in as_etree:
        material_tree = as_etree[material_key]
        for MaterialDomain in material_tree.iter("{*}MaterialDomain"):
            material_id = MaterialDomain.attrib["id"]
            material_file = MaterialDomain.attrib["associatedFile"].split("urn:3DXML:")[
                -1
            ]
            rend = as_etree[material_file].find("{*}Feature[@Alias='RenderingFeature']")
            diffuse = rend.find("{*}Attr[@Name='DiffuseColor']")
            # specular = rend.find("{*}Attr[@Name='SpecularColor']")
            # emissive = rend.find("{*}Attr[@Name='EmissiveColor']")
            if diffuse is not None:
                rgb = (np.array(json.loads(diffuse.attrib["Value"])) * 255).astype(
                    np.uint8
                )
                colors[material_id] = rgb
            texture = rend.find("{*}Attr[@Name='TextureImage']")
            if texture is not None:
                tex_file, tex_id = texture.attrib["Value"].split(":")[-1].split("#")
                rep_image = as_etree[tex_file].find(
                    f"{{*}}CATRepImage/{{*}}CATRepresentationImage[@id='{tex_id}']"
                )
                if rep_image is not None:
                    image_file = rep_image.get("associatedFile", "").split(":")[-1]
                    images[material_id] = Image.open(archive[image_file])

        # copy indexes for instances of colors
        for MaterialDomainInstance in material_tree.iter("{*}MaterialDomainInstance"):
            instance = MaterialDomainInstance.find("{*}IsInstanceOf")
            # colors[b.attrib['id']] = colors[instance.text]
            for aggregate in MaterialDomainInstance.findall("{*}IsAggregatedBy"):
                colors[aggregate.text] = colors.get(instance.text)
                images[aggregate.text] = images.get(instance.text)

    # references which hold the 3DXML scene structure as a dict
    # element id : {key : value}
    references = collections.defaultdict(dict)

    def get_rgba(color):
        """
        Return (4,) uint8 color array defined by Color element attributes.

        Parameters
        -----------
        color : lxml.Element
          Element containing RGBA colors.

        Returns
        -----------
        as_int : (4,) np.uint8
          Colors as uint8 RGBA.
        """
        assert "RGBAColorType" in color.attrib.values()
        # colors will be float 0.0 - 1.0
        rgba = np.array(
            [color.get(channel, 1.0) for channel in ("red", "green", "blue", "alpha")],
            dtype=np.float64,
        )
        # convert to int colors
        return (rgba * 255).astype(np.uint8)

    # the 3DXML can specify different visual properties for occurrences
    view = tree.find("{*}DefaultView")
    if view is not None:
        for ViewProp in view.iter("{*}DefaultViewProperty"):
            color = ViewProp.find(
                "{*}GraphicProperties/" + "{*}SurfaceAttributes/{*}Color"
            )
            if color is None:
                continue
            rgba = get_rgba(color)
            for occurrence in ViewProp.findall("{*}OccurenceId/{*}id"):
                reference_id = occurrence.text.split("#")[-1]
                references[reference_id]["color"] = rgba

    # geometries will hold meshes
    geometries = {}

    # get geometry
    for ReferenceRep in tree.iter(tag="{*}ReferenceRep"):
        # the str of an int that represents this meshes unique ID
        part_id = ReferenceRep.attrib["id"]
        # which part file in the archive contains the geometry we care about
        part_file = ReferenceRep.attrib["associatedFile"].split(":")[-1]
        # the format of the geometry file
        part_format = ReferenceRep.attrib["format"]
        if part_format not in ("TESSELLATED",):
            util.log.warning(
                f"ReferenceRep {part_file!r} unsupported format {part_format!r}"
            )
            continue

        # load actual geometry
        mesh_faces = []
        mesh_colors = []
        mesh_normals = []
        mesh_vertices = []
        mesh_uv = []
        mesh_image = None

        if part_file not in as_etree and part_file in archive:
            # the data is stored in some binary format
            util.log.warning(f"unable to load Rep {part_file!r}")
            # data = archive[part_file]
            continue

        # the geometry is stored in a Rep
        for Rep in as_etree[part_file].iter("{*}Rep"):
            rep_faces = []  # faces sharing the same list of vertices
            vertices = Rep.find("{*}VertexBuffer/{*}Positions")
            if vertices is None:
                continue

            # they mix delimiters like we couldn't figure it out from the
            # shape :(
            # load vertices into (n, 3) float64
            mesh_vertices.append(
                np.fromstring(
                    vertices.text.replace(",", " "), sep=" ", dtype=np.float64
                ).reshape((-1, 3))
            )

            # load vertex normals into (n, 3) float64
            normals = Rep.find("{*}VertexBuffer/{*}Normals")
            mesh_normals.append(
                np.fromstring(
                    normals.text.replace(",", " "), sep=" ", dtype=np.float64
                ).reshape((-1, 3))
            )

            uv = Rep.find("{*}VertexBuffer/{*}TextureCoordinates")
            if uv is not None:  # texture coordinates are available
                rep_uv = np.fromstring(
                    uv.text.replace(",", " "), sep=" ", dtype=np.float64
                )
                if "1D" == uv.get("dimension"):
                    mesh_uv.append(np.stack([rep_uv, np.zeros(len(rep_uv))], axis=1))
                else:  # 2D
                    mesh_uv.append(rep_uv.reshape(-1, 2))

            material = Rep.find(
                "{*}SurfaceAttributes/" + "{*}MaterialApplication/" + "{*}MaterialId"
            )
            if material is None:
                material_id = None
            else:
                (material_file, material_id) = (
                    material.attrib["id"].split("urn:3DXML:")[-1].split("#")
                )
                mesh_image = images.get(material_id)  # texture for this Rep, if any

            for faces in Rep.iter("{*}Faces"):
                triangles = []  # mesh triangles for this Faces element
                for face in faces.iter("{*}Face"):
                    # Each Face may have optional strips, triangles or fans attributes
                    if "strips" in face.attrib:
                        # triangle strips, sequence of arbitrary length lists
                        # np.fromstring is substantially faster than np.array(i.split())
                        # inside the list comprehension
                        strips = [
                            np.fromstring(i, sep=" ", dtype=np.int64)
                            for i in face.attrib["strips"].split(",")
                        ]
                        # convert strips to (m, 3) int triangles
                        triangles.extend(util.triangle_strips_to_faces(strips))

                    if "triangles" in face.attrib:
                        triangles.extend(
                            np.fromstring(
                                face.attrib["triangles"], sep=" ", dtype=np.int64
                            ).reshape((-1, 3))
                        )

                    if "fans" in face.attrib:
                        fans = [
                            np.fromstring(i, sep=" ", dtype=np.int64)
                            for i in face.attrib["fans"].split(",")
                        ]
                        # convert fans to (m, 3) int triangles
                        triangles.extend(util.triangle_fans_to_faces(fans))

                    rep_faces.extend(triangles)

                    # store the material information as (m, 3) uint8 FACE COLORS
                    faceColor = colors.get(material_id, [128, 128, 128])
                    # each Face may have its own color
                    colorElement = face.find("{*}SurfaceAttributes/{*}Color")
                    if colorElement is not None:
                        faceColor = get_rgba(colorElement)[:3]
                    mesh_colors.append(np.tile(faceColor, (len(triangles), 1)))
            mesh_faces.append(rep_faces)

        # save each mesh as the kwargs for a trimesh.Trimesh constructor
        # aka, a Trimesh object can be created with trimesh.Trimesh(**mesh)
        # this avoids needing trimesh- specific imports in this IO function
        mesh = {}
        (mesh["vertices"], mesh["faces"]) = util.append_faces(mesh_vertices, mesh_faces)
        mesh["vertex_normals"] = np.vstack(mesh_normals)
        if mesh_uv and mesh_image:
            mesh["visual"] = TextureVisuals(uv=np.vstack(mesh_uv), image=mesh_image)
        else:
            mesh["face_colors"] = np.vstack(mesh_colors)

        # as far as I can tell, all 3DXML files are exported as
        # implicit millimeters (it isn't specified in the file)
        mesh["metadata"] = {"units": "mm"}
        mesh["class"] = "Trimesh"

        geometries[part_id] = mesh
        references[part_id]["geometry"] = part_id

    # a Reference3D maps to a subassembly or assembly
    for Reference3D in tree.iter("{*}Reference3D"):
        references[Reference3D.attrib["id"]] = {
            "name": Reference3D.attrib["name"],
            "type": "Reference3D",
        }

    # a node that is the connectivity between a geometry and the Reference3D
    for InstanceRep in tree.iter("{*}InstanceRep"):
        current = InstanceRep.attrib["id"]
        instance = InstanceRep.find("{*}IsInstanceOf").text
        aggregate = InstanceRep.find("{*}IsAggregatedBy").text

        references[current].update(
            {"aggregate": aggregate, "instance": instance, "type": "InstanceRep"}
        )

    # an Instance3D maps basically to a part
    for Instance3D in tree.iter("{*}Instance3D"):
        matrix = np.eye(4)
        relative = Instance3D.find("{*}RelativeMatrix")
        if relative is not None:
            relative = np.array(relative.text.split(), dtype=np.float64)

            # rotation component
            matrix[:3, :3] = relative[:9].reshape((3, 3)).T
            # translation component
            matrix[:3, 3] = relative[9:]

        current = Instance3D.attrib["id"]
        name = Instance3D.attrib["name"]
        instance = Instance3D.find("{*}IsInstanceOf").text
        aggregate = Instance3D.find("{*}IsAggregatedBy").text

        references[current].update(
            {
                "aggregate": aggregate,
                "instance": instance,
                "matrix": matrix,
                "name": name,
                "type": "Instance3D",
            }
        )

    # turn references into directed graph for path finding
    graph = nx.DiGraph()
    for k, v in references.items():
        # IsAggregatedBy points up to a parent
        if "aggregate" in v:
            graph.add_edge(v["aggregate"], k)
        # IsInstanceOf indicates a child
        if "instance" in v:
            graph.add_edge(k, v["instance"])

    # the 3DXML format is stored as a directed acyclic graph that needs all
    # paths from the root to a geometry to generate the tree of the scene
    paths = []
    for geometry_id in geometries.keys():
        paths.extend(nx.all_simple_paths(graph, source=root_id, target=geometry_id))

    # the name of the root frame
    root_name = references[root_id]["name"]
    # create a list of kwargs to send to the scene.graph.update function
    # start with a transform from the graphs base frame to our root name

    graph_kwargs = [{"frame_to": root_name, "matrix": np.eye(4)}]

    # we are going to collect prettier geometry names as we traverse paths
    geom_names = {}
    # loop through every simple path and generate transforms tree
    # note that we are flattening the transform tree here
    for path in paths:
        name = ""
        if "name" in references[path[-3]]:
            name = references[path[-3]]["name"]
            geom_names[path[-1]] = name
        # we need a unique node name for our geometry instance frame
        # due to the nature of the DAG names specified by the file may not
        # be unique, so we add an Instance3D name then append the path ids
        node_name = name + "#" + ":".join(path)

        # pull all transformations in the path
        matrices = [references[i]["matrix"] for i in path if "matrix" in references[i]]
        if len(matrices) == 0:
            matrix = np.eye(4)
        elif len(matrices) == 1:
            matrix = matrices[0]
        else:
            matrix = util.multi_dot(matrices)

        graph_kwargs.append(
            {
                "matrix": matrix,
                "frame_from": root_name,
                "frame_to": node_name,
                "geometry": path[-1],
            }
        )

    # remap geometry names from id numbers to the name string
    # we extracted from the 3DXML tree
    geom_final = {}
    for key, value in geometries.items():
        if key in geom_names:
            geom_final[geom_names[key]] = value
    # change geometry names in graph kwargs in place
    for kwarg in graph_kwargs:
        if "geometry" not in kwarg:
            continue
        kwarg["geometry"] = geom_names[kwarg["geometry"]]

    # create the kwargs for load_kwargs
    result = {"class": "Scene", "geometry": geom_final, "graph": graph_kwargs}

    return result


def print_element(element):
    """
    Pretty-print an lxml.etree element.

    Parameters
    ------------
    element : etree element
    """
    pretty = etree.tostring(element, pretty_print=True).decode("utf-8")
    return pretty


try:
    # soft dependencies
    import networkx as nx
    from lxml import etree

    _threedxml_loaders = {"3dxml": load_3DXML}
except BaseException as E:
    # set loader to exception wrapper
    from ..exceptions import ExceptionWrapper

    _threedxml_loaders = {"3dxml": ExceptionWrapper(E)}



try:
    from . import generic as g
except BaseException:
    import generic as g


class SmoothTest(g.unittest.TestCase):
    def test_smooth(self):
        """
        Load a collada scene with pycollada.
        """
        m = g.trimesh.creation.icosahedron()
        m.vertices, m.faces = g.trimesh.remesh.subdivide_to_size(m.vertices, m.faces, 0.1)

        s = m.copy()
        q = m.copy()
        f = m.copy()
        d = m.copy()
        b = m.copy()
        v = m.copy()

        assert m.is_volume

        # Equal Weights
        lap = g.trimesh.smoothing.laplacian_calculation(mesh=m, equal_weight=True)

        g.trimesh.smoothing.filter_laplacian(s, 0.5, 10, False, True, lap)
        g.trimesh.smoothing.filter_laplacian(q, 0.5, 10, True, True, lap)
        g.trimesh.smoothing.filter_humphrey(f, 0.1, 0.5, 10, lap)
        g.trimesh.smoothing.filter_taubin(d, 0.5, 0.53, 10, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(b, 0.5, 10, False, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(v, 0.5, 10, True, lap)

        assert s.is_volume
        assert q.is_volume
        assert f.is_volume
        assert d.is_volume
        assert b.is_volume
        assert v.is_volume

        assert g.np.isclose(s.volume, m.volume, rtol=0.1)
        assert g.np.isclose(q.volume, m.volume, rtol=0.1)
        assert g.np.isclose(f.volume, m.volume, rtol=0.1)
        assert g.np.isclose(d.volume, m.volume, rtol=0.1)
        assert g.np.isclose(b.volume, m.volume, rtol=0.1)
        assert g.np.isclose(v.volume, m.volume, rtol=0.1)

        s = m.copy()
        q = m.copy()
        f = m.copy()
        d = m.copy()
        b = m.copy()
        v = m.copy()

        # umbrella Weights
        lap = g.trimesh.smoothing.laplacian_calculation(m, equal_weight=False)

        g.trimesh.smoothing.filter_laplacian(s, 0.5, 10, False, True, lap)
        g.trimesh.smoothing.filter_laplacian(q, 0.5, 10, True, True, lap)
        g.trimesh.smoothing.filter_humphrey(f, 0.1, 0.5, 10, lap)
        g.trimesh.smoothing.filter_taubin(d, 0.5, 0.53, 10, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(b, 0.5, 10, False, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(v, 0.5, 10, True, lap)

        assert s.is_volume
        assert q.is_volume
        assert f.is_volume
        assert d.is_volume
        assert b.is_volume
        assert v.is_volume

        assert g.np.isclose(s.volume, m.volume, rtol=0.1)
        assert g.np.isclose(q.volume, m.volume, rtol=0.1)
        assert g.np.isclose(f.volume, m.volume, rtol=0.1)
        assert g.np.isclose(d.volume, m.volume, rtol=0.1)
        assert g.np.isclose(b.volume, m.volume, rtol=0.1)
        assert g.np.isclose(v.volume, m.volume, rtol=0.1)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

try:
    import collada
except ImportError:
    collada = None
except BaseException:
    # TODO : REMOVE WHEN UPSTREAM RELEASE FIXED
    # https://github.com/pycollada/pycollada/pull/92
    g.log.error("DAE fix not pushed yet!")
    collada = None


class DAETest(g.unittest.TestCase):
    def test_duck(self):
        """
        Load a collada scene with pycollada.
        """
        if collada is None:
            g.log.error("no pycollada to test!")
            return

        scene = g.get_mesh("duck.dae")

        assert len(scene.geometry) == 1
        assert len(scene.graph.nodes_geometry) == 1

        conv = scene.convert_units("inch")
        assert conv.units == "inch"

    def test_shoulder(self):
        if collada is None:
            g.log.error("no pycollada to test!")
            return

        scene = g.get_mesh("shoulder.zae")
        assert len(scene.geometry) == 3
        assert len(scene.graph.nodes_geometry) == 3

        assert scene.units != "mm"
        conv = scene.convert_units("mm")
        assert conv.units == "mm"

    def test_export(self):
        if collada is None:
            g.log.error("no pycollada to test!")
            return

        a = g.get_mesh("ballA.off")
        r = a.export(file_type="dae")
        assert len(r) > 0

    def test_obj_roundtrip(self):
        # get a zipped-DAE scene
        s = g.get_mesh("duck.zae", force="mesh")
        with g.TemporaryDirectory() as root:
            # export using a file path so it can auto-create
            # a FilePathResolver to write the stupid assets
            path = g.os.path.join(root, "duck.obj")
            s.export(path)
            # bring it back from outer space
            rec = g.trimesh.load(path, force="mesh")
        assert g.np.ptp(g.np.ptp(rec.visual.uv, axis=0)) > 1e-5
        assert s.visual.material.baseColorTexture.size == rec.visual.material.image.size

        conv = s.convert_units("inch")
        assert conv.units == "inch"

    def test_material_round(self):
        """
        Test to make sure materials survive a roundtrip
        with an actually identical result
        """
        s = g.get_mesh("blue_cube.dae")
        assert len(s.geometry) == 1
        m = next(iter(s.geometry.values()))

        rs = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(m.export(file_type="dae")),
            file_type="dae",
        )
        assert len(rs.geometry) == 1
        r = next(iter(rs.geometry.values()))

        # this will compare everything in `material._data`
        assert hash(m.visual.material) == hash(r.visual.material)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
xaml.py
---------

Load 3D XAMl files, an export option from Solidworks.
"""

import collections

import numpy as np

from .. import transformations as tf
from .. import util, visual


def load_XAML(file_obj, *args, **kwargs):
    """
    Load a 3D XAML file.

    Parameters
    ----------
    file_obj : file object
      Open XAML file.

    Returns
    ----------
    result : dict
      Kwargs for a Trimesh constructor.
    """

    def element_to_color(element):
        """
        Turn an XML element into a (4,) np.uint8 RGBA color
        """
        if element is None:
            return visual.DEFAULT_COLOR
        hexcolor = int(element.attrib["Color"].replace("#", ""), 16)
        opacity = float(element.attrib["Opacity"])
        rgba = [
            (hexcolor >> 16) & 0xFF,
            (hexcolor >> 8) & 0xFF,
            (hexcolor & 0xFF),
            opacity * 0xFF,
        ]
        rgba = np.array(rgba, dtype=np.uint8)
        return rgba

    def element_to_transform(element):
        """
        Turn an XML element into a (4,4) np.float64
        transformation matrix.
        """
        try:
            matrix = next(element.iter(tag=ns + "MatrixTransform3D")).attrib["Matrix"]
            matrix = np.array(matrix.split(), dtype=np.float64).reshape((4, 4)).T
            return matrix
        except StopIteration:
            # this will be raised if the MatrixTransform3D isn't in the passed
            # elements tree
            return np.eye(4)

    # read the file and parse XML
    file_data = file_obj.read()
    root = etree.XML(file_data)

    # the XML namespace
    ns = root.tag.split("}")[0] + "}"

    # the linked lists our results are going in
    vertices = []
    faces = []
    colors = []
    normals = []

    # iterate through the element tree
    # the GeometryModel3D tag contains a material and geometry
    for geometry in root.iter(tag=ns + "GeometryModel3D"):
        # get the diffuse and specular colors specified in the material
        color_search = ".//{ns}{color}Material/*/{ns}SolidColorBrush"
        diffuse = geometry.find(color_search.format(ns=ns, color="Diffuse"))
        specular = geometry.find(color_search.format(ns=ns, color="Specular"))

        # convert the element into a (4,) np.uint8 RGBA color
        diffuse = element_to_color(diffuse)
        specular = element_to_color(specular)

        # to get the final transform of a component we'll have to traverse
        # all the way back to the root node and save transforms we find
        current = geometry
        transforms = collections.deque()
        # when the root node is reached its parent will be None and we stop
        while current is not None:
            # element.find will only return elements that are direct children
            # of the current element as opposed to element.iter,
            # which will return any depth of child
            transform_element = current.find(ns + "ModelVisual3D.Transform")
            if transform_element is not None:
                # we are traversing the tree backwards, so append new
                # transforms to the left of the deque
                transforms.appendleft(element_to_transform(transform_element))
            # we are going from the lowest level of the tree to the highest
            # this avoids having to traverse any branches that don't have
            # geometry
            current = current.getparent()

        if len(transforms) == 0:
            # no transforms in the tree mean an identity matrix
            transform = np.eye(4)
        elif len(transforms) == 1:
            # one transform in the tree we can just use
            transform = transforms.pop()
        else:
            # multiple transforms we apply all of them in order
            transform = util.multi_dot(transforms)

        # iterate through the contained mesh geometry elements
        for g in geometry.iter(tag=ns + "MeshGeometry3D"):
            c_normals = np.array(
                g.attrib["Normals"].replace(",", " ").split(), dtype=np.float64
            ).reshape((-1, 3))

            c_vertices = np.array(
                g.attrib["Positions"].replace(",", " ").split(), dtype=np.float64
            ).reshape((-1, 3))
            # bake in the transform as we're saving
            c_vertices = tf.transform_points(c_vertices, transform)

            c_faces = np.array(
                g.attrib["TriangleIndices"].replace(",", " ").split(), dtype=np.int64
            ).reshape((-1, 3))

            # save data to a sequence
            vertices.append(c_vertices)
            faces.append(c_faces)
            colors.append(np.tile(diffuse, (len(c_faces), 1)))
            normals.append(c_normals)

    # compile the results into clean numpy arrays
    result = {}
    result["vertices"], result["faces"] = util.append_faces(vertices, faces)
    result["face_colors"] = np.vstack(colors)
    result["vertex_normals"] = np.vstack(normals)

    return result


try:
    from lxml import etree

    _xaml_loaders = {"xaml": load_XAML}
except BaseException as E:
    # create a dummy module which will raise the ImportError
    # or other exception only when someone tries to use networkx
    from ..exceptions import ExceptionWrapper

    _xaml_loaders = {"xaml": ExceptionWrapper(E)}



try:
    from . import generic as g
except BaseException:
    import generic as g


class IntegralMeanCurvatureTest(g.unittest.TestCase):
    def test_IMCsphere(self):
        # how close do we need to be - relative tolerance
        tol = 1e-3
        for radius in [0.1, 1.0, 3.1459, 29.20]:
            m = g.trimesh.creation.icosphere(subdivisions=4, radius=radius)
            IMC = m.integral_mean_curvature
            ref = 4 * g.np.pi * radius
            assert g.np.isclose(IMC, ref, rtol=tol)

    def test_IMCcapsule(self):
        # how close do we need to be - relative tolerance
        tol = 1e-3
        radius = 1.2
        for aspect_ratio in [0.0, 0.5, 1.0, 4.0, 100]:
            L = aspect_ratio * radius
            m = g.trimesh.creation.capsule(height=L, radius=radius, count=[64, 64])
            IMC = m.integral_mean_curvature
            ref = g.np.pi * (L + 4 * radius)
            assert g.np.isclose(IMC, ref, rtol=tol)

    def test_IMCbox(self):
        # how close do we need to be - relative tolerance
        tol = 1e-3
        n_tests = 4
        extents = 1 - g.random((n_tests, 3))
        for extent in extents:
            m = g.trimesh.creation.box(extents=extent)
            IMC = m.integral_mean_curvature
            # only the right angle (=pi/2) edges contribute,
            # edges that lie on the faces of the box do not
            ref = 4 * extent.sum() * g.np.pi / 2 * 0.5
            assert g.np.isclose(IMC, ref, rtol=tol)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class FillTest(g.unittest.TestCase):
    def test_fill(self):
        # a path closed with a bowtie, so the topology is wrong
        a = g.get_mesh("2D/broken_loop.dxf")
        assert len(a.paths) == 0
        # bowtie shouldn't require any connection distance
        a.fill_gaps(0.0)
        # topology should be good now
        assert len(a.paths) == 1
        # it is a rectangle
        assert g.np.isclose(a.area, g.np.prod(a.extents))

        # a path with a bowtie and a .05 gap
        b = g.get_mesh("2D/broken_pair.dxf")
        assert len(b.paths) == 0
        # should be too small to fill gap
        b.fill_gaps(0.01)
        assert len(b.paths) == 0
        # should be large enough to fill gap
        b.fill_gaps(0.06)
        assert len(b.paths) == 1
        # it is a rectangle
        assert g.np.isclose(b.area, g.np.prod(b.extents))


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class UnitsTest(g.unittest.TestCase):
    def test_units(self):
        # make sure unit conversions fail for fake units
        fake_units = "blorbs"
        fake_units = "in"
        try:
            c = g.trimesh.units.unit_conversion("inches", fake_units)  # NOQA
            raise AssertionError()
        except BaseException:
            pass

        m = g.get_mesh("featuretype.STL")
        self.assertTrue(m.units is None)

        m.units = "in"
        self.assertTrue(m.units == "in")

        extents_pre = m.extents
        m.convert_units("mm")
        scale = g.np.divide(m.extents, extents_pre)
        self.assertTrue(g.np.allclose(scale, 25.4))
        self.assertTrue(m.units == "mm")

    def test_conversion(self):
        # test conversions on a multibody STL in a scene

        # a multibody STL with a unit hint in filename
        m = g.get_mesh("counter.unitsmm.STL")

        # nothing should be set
        assert m.units is None

        # split into watertight bodies
        s = g.trimesh.scene.split_scene(m)

        # save the extents
        extents_pre = s.extents

        # should extract units from file name without
        # raising a ValueError
        c = s.convert_units("in", guess=False)
        # should have converted mm -> in, 1/25.4
        # extents should scale exactly with unit conversion
        assert g.np.allclose(extents_pre / c.extents, 25.4, atol=0.01)

    def test_path(self):
        p = g.get_mesh("2D/tray-easy1.dxf")
        # should be inches
        assert "in" in p.units
        extents_pre = p.extents
        p.convert_units("mm")
        # should have converted in -> mm 25.4
        # extents should scale exactly with unit conversion
        assert g.np.allclose(p.extents / extents_pre, 25.4, atol=0.01)

    def test_keys(self):
        units = g.trimesh.units.keys()
        assert isinstance(units, set)
        assert "in" in units

    def test_arbitrary(self):
        ac = g.np.allclose
        to_inch = g.trimesh.units.to_inch

        # check whitespace
        assert ac(1.0, to_inch("in"))
        assert ac(1.0, to_inch("1.00000* in"))
        assert ac(1.0, to_inch("1.00    * in"))

        # check centimeter conversion
        assert ac(100, to_inch("m") / to_inch("0.01*m"))

        # if we are currently in centimeters and want to go to meters
        # it should be dividing it by 100
        assert ac(
            0.01, g.trimesh.units.unit_conversion(current="0.01*m", desired="meters")
        )


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import re
from collections import defaultdict, deque

import numpy as np

try:
    # `pip install pillow`
    # optional: used for textured meshes
    from PIL import Image
except BaseException as E:
    # if someone tries to use Image re-raise
    # the import error so they can debug easily
    from ..exceptions import ExceptionWrapper

    Image = ExceptionWrapper(E)

from .. import util
from ..constants import log, tol
from ..visual.color import to_float
from ..visual.material import SimpleMaterial
from ..visual.texture import TextureVisuals, unmerge_faces


def load_obj(
    file_obj,
    resolver=None,
    group_material=True,
    skip_materials=False,
    maintain_order=False,
    **kwargs,
):
    """
    Load a Wavefront OBJ file into kwargs for a trimesh.Scene
    object.

    Parameters
    --------------
    file_obj : file like object
      Contains OBJ data
    resolver : trimesh.visual.resolvers.Resolver
      Allow assets such as referenced textures and
      material files to be loaded
    group_material : bool
      Group faces that share the same material
      into the same mesh.
    skip_materials : bool
      Don't load any materials.
    maintain_order : bool or None
      Do not reorder faces or vertices which may result
      in visual artifacts.

    Returns
    -------------
    kwargs : dict
      Keyword arguments which can be loaded by
      trimesh.exchange.load.load_kwargs into a trimesh.Scene
    """
    # get text as bytes or string blob
    text = file_obj.read()
    # if text was bytes decode into string
    text = util.decode_text(text)

    # add leading and trailing newlines so we can use the
    # same logic even if they jump directly in to data lines
    text = "\n{}\n".format(text.strip().replace("\r\n", "\n"))

    # remove backslash continuation characters and merge them into the same
    # line
    text = text.replace("\\\n", "")

    # Load Materials
    materials = {}
    mtl_position = text.find("mtllib")
    if not skip_materials and mtl_position >= 0:
        # take the line of the material file after `mtllib`
        # which should be the file location of the .mtl file
        mtl_path = text[mtl_position + 6 : text.find("\n", mtl_position)].strip()
        try:
            # use the resolver to get the data
            material_kwargs = parse_mtl(resolver[mtl_path], resolver=resolver)
            # turn parsed kwargs into material objects
            materials = {k: SimpleMaterial(**v) for k, v in material_kwargs.items()}
        except (OSError, TypeError):
            # usually the resolver couldn't find the asset
            log.debug(f"unable to load materials from: {mtl_path}")
        except BaseException:
            # something else happened so log a warning
            log.debug(f"unable to load materials from: {mtl_path}", exc_info=True)

    # extract vertices from raw text
    v, vn, vt, vc = _parse_vertices(text=text)

    # get relevant chunks that have face data
    # in the form of (material, object, chunk)
    face_tuples = _preprocess_faces(text=text)

    # combine chunks that have the same material
    # some meshes end up with a LOT of components
    # and will be much slower if you don't do this
    if group_material:
        face_tuples = _group_by_material(face_tuples)

    # no faces but points given
    # return point cloud
    if not len(face_tuples) and v is not None:
        pc = {"vertices": v}
        if vn is not None:
            pc["vertex_normals"] = vn
        if vc is not None:
            pc["vertex_colors"] = vc
        return pc

    # Load Faces
    # now we have clean- ish faces grouped by material and object
    # so now we have to turn them into numpy arrays and kwargs
    # for trimesh mesh and scene objects
    geometry = {}
    while len(face_tuples) > 0:
        # consume the next chunk of text
        material, current_object, chunk = face_tuples.pop()
        # do wangling in string form
        # we need to only take the face line before a newline
        # using builtin functions in a list comprehension
        # is pretty fast relative to other options
        # this operation is the only one that is O(len(faces))
        # slower due to the tight-loop conditional:
        # face_lines = [i[:i.find('\n')]
        #              for i in chunk.split('\nf ')[1:]
        #              if i.rfind('\n') >0]
        # maxsplit=1 means that it can stop working
        # after it finds the first newline
        # passed as arg as it's not a kwarg in python2
        face_lines = [
            i.split("\n", 1)[0].strip()
            for i in re.split("^f", chunk, flags=re.MULTILINE)[1:]
        ]

        # check every face for mixed tri-quad-ngon
        columns = len(face_lines[0].replace("/", " ").split())
        flat_array = all(columns == len(f.replace("/", " ").split()) for f in face_lines)

        # make sure we have the right number of values for vectorized
        if flat_array:
            # the fastest way to get to a numpy array
            # processes the whole string at once into a 1D array
            array = np.fromstring(
                " ".join(face_lines).replace("/", " "), sep=" ", dtype=np.int64
            )
            # also wavefront is 1-indexed (vs 0-indexed) so offset
            # only applies to positive indices
            array[array > 0] -= 1

            # everything is a nice 2D array
            faces, faces_tex, faces_norm = _parse_faces_vectorized(
                array=array, columns=columns, sample_line=face_lines[0]
            )
        else:
            # if we had something annoying like mixed in quads
            # or faces that differ per-line we have to loop
            # i.e. something like:
            #  '31407 31406 31408',
            #  '32303/2469 32304/2469 32305/2469',
            log.debug("faces have mixed data: using slow fallback!")
            faces, faces_tex, faces_norm = _parse_faces_fallback(face_lines)

        if group_material:
            name = material
        else:
            name = current_object

        # ensure the name is always unique
        name = util.unique_name(name, geometry)

        # try to get usable texture
        mesh = kwargs.copy()
        if faces_tex is not None:
            # convert faces referencing vertices and
            # faces referencing vertex texture to new faces
            # where each face
            if faces_norm is not None and len(faces_norm) == len(faces):
                new_faces, mask_v, mask_vt, mask_vn = unmerge_faces(
                    faces, faces_tex, faces_norm, maintain_faces=maintain_order
                )
            else:
                mask_vn = None
                # no face normals but face texturre
                new_faces, mask_v, mask_vt = unmerge_faces(
                    faces, faces_tex, maintain_faces=maintain_order
                )

            if tol.strict:
                # we should NOT have messed up the faces
                # note: this is EXTREMELY slow due to all the
                # float comparisons so only run this in unit tests
                assert np.allclose(v[faces], v[mask_v][new_faces])
                # faces should all be in bounds of vertives
                assert new_faces.max() < len(v[mask_v])
            try:
                # survive index errors as sometimes we
                # want materials without UV coordinates
                uv = vt[mask_vt]
            except BaseException:
                log.debug("index failed on UV coordinates, skipping!")
                uv = None

            # mask vertices and use new faces
            mesh.update({"vertices": v[mask_v].copy(), "faces": new_faces})

        else:
            # otherwise just use unmasked vertices
            uv = None
            # check to make sure indexes are in bounds
            if tol.strict:
                assert faces.max() < len(v)
            if vn is not None and np.shape(faces_norm) == faces.shape:
                # do the crazy unmerging logic for split indices
                new_faces, mask_v, mask_vn = unmerge_faces(
                    faces, faces_norm, maintain_faces=maintain_order
                )
            else:
                # generate the mask so we only include
                # referenced vertices in every new mesh
                mask_v = np.zeros(len(v), dtype=bool)
                mask_v[faces] = True

                # reconstruct the faces with the new vertex indices
                inverse = np.zeros(len(v), dtype=np.int64)
                inverse[mask_v] = np.arange(mask_v.sum())
                new_faces = inverse[faces]
                # no normals
                mask_vn = None

            # start with vertices and faces
            mesh.update({"faces": new_faces, "vertices": v[mask_v].copy()})

        # if colors and normals are OK save them
        if vc is not None:
            try:
                # may fail on a malformed color mask
                mesh["vertex_colors"] = vc[mask_v]
            except BaseException:
                log.debug("failed to load vertex_colors", exc_info=True)
        if mask_vn is not None:
            try:
                # may fail on a malformed mask
                normals = vn[mask_vn]
                if normals.shape != mesh["vertices"].shape:
                    raise ValueError(
                        "incorrect normals {} != {}".format(
                            str(normals.shape), str(mesh["vertices"].shape)
                        )
                    )
                mesh["vertex_normals"] = normals
            except BaseException:
                log.debug("failed to load vertex_normals", exc_info=True)

        visual = None
        if material in materials:
            # use the material with the UV coordinates
            visual = TextureVisuals(uv=uv, material=materials[material])
        elif uv is not None and len(uv) == len(mesh["vertices"]):
            # create a texture with an empty materials
            visual = TextureVisuals(uv=uv)
        elif material is not None:
            # case where material is specified but not available
            log.debug(f"specified material ({material})  not loaded!")
        # assign the visual
        mesh["visual"] = visual
        # store geometry by name
        geometry[name] = mesh

    if len(geometry) == 1:
        # TODO : should this be removed to always return a scene?
        return next(iter(geometry.values()))

    # add an identity transform for every geometry
    graph = [{"geometry": k, "frame_to": k} for k in geometry.keys()]

    # convert to scene kwargs
    result = {"geometry": geometry, "graph": graph}

    return result


def parse_mtl(mtl, resolver=None):
    """
    Parse a loaded MTL file.

    Parameters
    -------------
    mtl : str or bytes
      Data from an MTL file
    resolver : trimesh.Resolver
      Fetch assets by name from file system, web, or other

    Returns
    ------------
    mtllibs : list of dict
      Each dict has keys: newmtl, map_Kd, Kd
    """
    # decode bytes into string if necessary
    mtl = util.decode_text(mtl)

    # current material
    material = None
    # materials referenced by name
    materials = {}
    # use universal newline splitting
    lines = str.splitlines(str(mtl).strip())

    # remap OBJ property names to kwargs for SimpleMaterial
    mapped = {"kd": "diffuse", "ka": "ambient", "ks": "specular", "ns": "glossiness"}

    for line in lines:
        # split by white space
        split = line.strip().split()
        # needs to be at least two values
        if len(split) <= 1:
            continue
        # the first value is the parameter name
        key = split[0].lower()
        # start a new material
        if key == "newmtl":
            # material name extracted from line like:
            # newmtl material_0
            if material is not None:
                # save the old material by old name and remove key
                materials[material.pop("newmtl")] = material
            # start a fresh new material
            material = {"newmtl": " ".join(split[1:])}

        elif key == "map_kd":
            # represents the file name of the texture image
            index = line.lower().index("map_kd") + 6
            file_name = line[index:].strip()
            try:
                file_data = resolver.get(file_name)
                # load the bytes into a PIL image
                # an image file name
                material["image"] = Image.open(util.wrap_as_stream(file_data))
            except BaseException:
                log.debug("failed to load image", exc_info=True)

        elif key in mapped.keys():
            try:
                # diffuse, ambient, and specular float RGB
                value = [float(x) for x in split[1:]]
                # if there is only one value return that
                if len(value) == 1:
                    value = value[0]
                # store the key by mapped value
                material[mapped[key]] = value
                # also store key by OBJ name
                material[key] = value
            except BaseException:
                log.debug("failed to convert color!", exc_info=True)
        # pass everything as kwargs to material constructor
        elif material is not None:
            # save any other unspecified keys
            material[key] = split[1:]
    # reached EOF so save any existing materials
    if material:
        materials[material.pop("newmtl")] = material

    return materials


def _parse_faces_vectorized(array, columns, sample_line):
    """
    Parse loaded homogeneous (tri/quad) face data in a
    vectorized manner.

    Parameters
    ------------
    array : (n,) int
      Indices in order
    columns : int
      Number of columns in the file
    sample_line : str
      A single line so we can assess the ordering

    Returns
    --------------
    faces : (n, d) int
      Faces in space
    faces_tex : (n, d) int or None
      Texture for each vertex in face
    faces_norm : (n, d) int or None
      Normal index for each vertex in face
    """
    # reshape to columns
    array = array.reshape((-1, columns))
    # how many elements are in the first line of faces
    # i.e '13/1/13 14/1/14 2/1/2 1/2/1' is 4
    group_count = len(sample_line.strip().split())
    # how many elements are there for each vertex reference
    # i.e. '12/1/13' is 3
    per_ref = int(columns / group_count)
    # create an index mask we can use to slice vertex references
    index = np.arange(group_count) * per_ref
    # slice the faces out of the blob array
    faces = array[:, index]

    # TODO: probably need to support 8 and 12 columns for quads
    # or do something more general
    faces_tex, faces_norm = None, None
    if columns == 6:
        # if we have two values per vertex the second
        # one is index of texture coordinate (`vt`)
        # count how many delimiters are in the first face line
        # to see if our second value is texture or normals
        # do splitting to clip off leading/trailing slashes
        count = "".join(i.strip("/") for i in sample_line.split()).count("/")
        if count == columns:
            # case where each face line looks like:
            # ' 75//139 76//141 77//141'
            # which is vertex/nothing/normal
            faces_norm = array[:, index + 1]
        elif count == int(columns / 2):
            # case where each face line looks like:
            # '75/139 76/141 77/141'
            # which is vertex/texture
            faces_tex = array[:, index + 1]
        else:
            log.debug(f"face lines are weird: {sample_line}")
    elif columns == 9:
        # if we have three values per vertex
        # second value is always texture
        faces_tex = array[:, index + 1]
        # third value is reference to vertex normal (`vn`)
        faces_norm = array[:, index + 2]
    return faces, faces_tex, faces_norm


def _parse_faces_fallback(lines):
    """
    Use a slow but more flexible looping method to process
    face lines as a fallback option to faster vectorized methods.

    Parameters
    -------------
    lines : (n,) str
      List of lines with face information

    Returns
    -------------
    faces : (m, 3) int
      Clean numpy array of face triangles
    """

    # collect vertex, texture, and vertex normal indexes
    v, vt, vn = [], [], []

    # loop through every line starting with a face
    for line in lines:
        # remove leading newlines then
        # take first bit before newline then split by whitespace
        split = line.strip().split("\n")[0].split()
        # split into: ['76/558/76', '498/265/498', '456/267/456']
        len_split = len(split)
        if len_split == 3:
            pass
        elif len_split == 4:
            # triangulate quad face
            split = [split[0], split[1], split[2], split[2], split[3], split[0]]
        elif len_split > 4:
            # triangulate polygon as a triangles fan
            collect = []
            # we need a flat list so append inside
            # a list comprehension
            collect_append = collect.append
            [
                [
                    collect_append(split[0]),
                    collect_append(split[i + 1]),
                    collect_append(split[i + 2]),
                ]
                for i in range(len(split) - 2)
            ]
            split = collect
        else:
            log.debug(f"face needs more values 3>{len(split)} skipping!")
            continue

        # f is like: '76/558/76'
        for f in split:
            # vertex, vertex texture, vertex normal
            split = f.split("/")
            # we always have a vertex reference
            v.append(int(split[0]))

            # faster to try/except than check in loop
            try:
                vt.append(int(split[1]))
            except BaseException:
                pass
            try:
                # vertex normal is the third index
                vn.append(int(split[2]))
            except BaseException:
                pass

    # shape into triangles and switch to 0-indexed
    # 0-indexing only applies to positive indices
    faces = np.array(v, dtype=np.int64).reshape((-1, 3))
    faces[faces > 0] -= 1
    faces_tex, normals = None, None
    if len(vt) == len(v):
        faces_tex = np.array(vt, dtype=np.int64).reshape((-1, 3))
        faces_tex[faces_tex > 0] -= 1
    if len(vn) == len(v):
        normals = np.array(vn, dtype=np.int64).reshape((-1, 3))
        normals[normals > 0] -= 1

    return faces, faces_tex, normals


def _parse_vertices(text):
    """
    Parse raw OBJ text into vertices, vertex normals,
    vertex colors, and vertex textures.

    Parameters
    -------------
    text : str
      Full text of an OBJ file

    Returns
    -------------
    v : (n, 3) float
      Vertices in space
    vn : (m, 3) float or None
      Vertex normals
    vt : (p, 2) float or None
      Vertex texture coordinates
    vc : (n, 3) float or None
      Per-vertex color
    """

    # the first position of a vertex in the text blob
    # we only really need to search from the start of the file
    # up to the location of out our first vertex but we
    # are going to use this check for "do we have texture"
    # determination later so search the whole stupid file
    starts = {k: text.find(f"\n{k} ") for k in ["v", "vt", "vn"]}

    # no valid values so exit early
    if not any(v >= 0 for v in starts.values()):
        return None, None, None, None

    # find the last position of each valid value
    ends = {
        k: text.find("\n", text.rfind(f"\n{k} ") + 2 + len(k))
        for k, v in starts.items()
        if v >= 0
    }

    # take the first and last position of any vertex property
    start = min(s for s in starts.values() if s >= 0)
    end = max(e for e in ends.values() if e >= 0)
    # get the chunk of test that contains vertex data
    chunk = text[start:end].replace("+e", "e").replace("-e", "e")

    # get the clean-ish data from the file as python lists
    data = {
        k: [i.split("\n", 1)[0] for i in chunk.split(f"\n{k} ")[1:]]
        for k, v in starts.items()
        if v >= 0
    }

    # count the number of data values per row on a sample row
    per_row = {k: len(v[0].split()) for k, v in data.items()}

    # convert data values into numpy arrays
    result = defaultdict(lambda: None)
    for k, value in data.items():
        # use joining and fromstring to get as numpy array
        array = np.fromstring(" ".join(value), sep=" ", dtype=np.float64)
        # what should our shape be
        shape = (len(value), per_row[k])
        # check shape of flat data
        if len(array) == np.prod(shape):
            # we have a nice 2D array
            result[k] = array.reshape(shape)
        else:
            # try to recover with a slightly more expensive loop
            count = per_row[k]
            try:
                # try to get result through reshaping
                result[k] = np.fromstring(
                    " ".join(i.split()[:count] for i in value), sep=" ", dtype=np.float64
                ).reshape(shape)
            except BaseException:
                pass

    # vertices
    v = result["v"]
    # vertex colors are stored next to vertices
    vc = None
    if v is not None and v.shape[1] >= 6:
        # vertex colors are stored after vertices
        v, vc = v[:, :3], v[:, 3:6]
    elif v is not None and v.shape[1] > 3:
        # we got a lot of something unknowable
        v = v[:, :3]

    # vertex texture or None
    vt = result["vt"]
    if vt is not None:
        # sometimes UV coordinates come in as UVW
        vt = vt[:, :2]
    # vertex normals or None
    vn = result["vn"]

    # check will generally only be run in unit tests
    # so we are allowed to do things that are slow
    if tol.strict:
        # check to make sure our subsetting
        # didn't miss any vertices or data
        assert len(v) == text.count("\nv ")
        # make sure optional data matches file too
        if vn is not None:
            assert len(vn) == text.count("\nvn ")
        if vt is not None:
            assert len(vt) == text.count("\nvt ")

    return v, vn, vt, vc


def _group_by_material(face_tuples):
    """
    For chunks of faces split by material group
    the chunks that share the same material.

    Parameters
    ------------
    face_tuples : (n,) list of (material, obj, chunk)
      The data containing faces

    Returns
    ------------
    grouped : (m,) list of (material, obj, chunk)
      Grouped by material
    """

    # store the chunks grouped by material
    grouped = defaultdict(lambda: ["", "", []])
    # loop through existring
    for material, obj, chunk in face_tuples:
        grouped[material][0] = material
        grouped[material][1] = obj
        # don't do a million string concatenations in loop
        grouped[material][2].append(chunk)
    # go back and do a join to make a single string
    for k in grouped.keys():
        grouped[k][2] = "\n".join(grouped[k][2])
    # return as list
    return list(grouped.values())


def _preprocess_faces(text):
    """
    Pre-Process Face Text

    Rather than looking at each line in a loop we're
    going to split lines by directives which indicate
    a new mesh, specifically 'usemtl' and 'o' keys
    search for materials, objects, faces, or groups

    Parameters
    ------------
    text : str
      Raw file

    Returns
    ------------
    triple : (n, 3) tuple
      Tuples of (material, object, data-chunk)
    """
    # see which chunk is relevant
    starters = ["\nusemtl ", "\no ", "\nf ", "\ng ", "\ns "]
    f_start = len(text)
    # first index of material, object, face, group, or smoother
    for st in starters:
        search = text.find(st, 0, f_start)
        # if not contained find will return -1
        if search < 0:
            continue
        # subtract the length of the key from the position
        # to make sure it's included in the slice of text
        if search < f_start:
            f_start = search
    # index in blob of the newline after the last face
    f_end = text.find("\n", text.rfind("\nf ") + 3)
    # get the chunk of the file that has face information
    if f_end >= 0:
        # clip to the newline after the last face
        f_chunk = text[f_start:f_end]
    else:
        # no newline after last face
        f_chunk = text[f_start:]

    if tol.strict:
        # check to make sure our subsetting didn't miss any faces
        assert f_chunk.count("\nf ") == text.count("\nf ")

    # two things cause new meshes to be created:
    # objects and materials
    # re.finditer was faster than find in a loop
    # find the index of every material change
    idx_mtl = np.array([m.start(0) for m in re.finditer("usemtl ", f_chunk)], dtype=int)
    # find the index of every new object
    idx_obj = np.array([m.start(0) for m in re.finditer("\no ", f_chunk)], dtype=int)

    # find all the indexes where we want to split
    splits = np.unique(np.concatenate(([0, len(f_chunk)], idx_mtl, idx_obj)))

    # track the current material and object ID
    current_obj = None
    current_mtl = None
    # store (material, object, face lines)
    face_tuples = []

    for start, end in zip(splits[:-1], splits[1:]):
        # ensure there's always a trailing newline
        chunk = f_chunk[start:end].strip() + "\n"
        if chunk.startswith("o "):
            current_obj, chunk = chunk.split("\n", 1)
            current_obj = current_obj[2:].strip()
        elif chunk.startswith("usemtl"):
            current_mtl, chunk = chunk.split("\n", 1)
            current_mtl = current_mtl[6:].strip()
        # Discard the g tag line in the list of faces
        elif chunk.startswith("g "):
            _, chunk = chunk.split("\n", 1)
        # If we have an f at the beginning of a line
        # then add it to the list of faces chunks
        if chunk.startswith("f ") or "\nf" in chunk:
            face_tuples.append((current_mtl, current_obj, chunk))
    return face_tuples


def export_obj(
    mesh,
    include_normals=None,
    include_color=True,
    include_texture=True,
    return_texture=False,
    write_texture=True,
    resolver=None,
    digits=8,
    mtl_name=None,
    header="https://github.com/mikedh/trimesh",
):
    """
    Export a mesh as a Wavefront OBJ file.
    TODO: scenes with textured meshes

    Parameters
    -----------
    mesh : trimesh.Trimesh
      Mesh to be exported
    include_normals : Optional[bool]
      Include vertex normals in export. If None
      will only be included if vertex normals are in cache.
    include_color : bool
      Include vertex color in export
    include_texture : bool
      Include `vt` texture in file text
    return_texture : bool
      If True, return a dict with texture files
    write_texture : bool
      If True and a writable resolver is passed
      write the referenced texture files with resolver
    resolver : None or trimesh.resolvers.Resolver
      Resolver which can write referenced text objects
    digits : int
      Number of digits to include for floating point
    mtl_name : None or str
      If passed, the file name of the MTL file.
    header : str or None
      Header string for top of file or None for no header.

    Returns
    -----------
    export : str
      OBJ format output
    texture : dict
      Contains files that need to be saved in the same
      directory as the exported mesh: {file name : bytes}
    """
    # store the multiple options for formatting
    # vertex indexes for faces
    face_formats = {
        ("v",): "{}",
        ("v", "vn"): "{}//{}",
        ("v", "vt"): "{}/{}",
        ("v", "vn", "vt"): "{}/{}/{}",
    }

    # check the input
    if util.is_instance_named(mesh, "Trimesh"):
        meshes = [mesh]
    elif util.is_instance_named(mesh, "Scene"):
        meshes = mesh.dump()
    elif util.is_instance_named(mesh, "PointCloud"):
        meshes = [mesh]
    else:
        raise ValueError("must be Trimesh or Scene!")

    # collect lines to export
    objects = deque([])
    # keep track of the number of each export element
    counts = {"v": 0, "vn": 0, "vt": 0}
    # collect materials as we go
    materials = {}
    materials_name = set()

    for current in meshes:
        # we are going to reference face_formats with this
        face_type = ["v"]
        # OBJ includes vertex color as RGB elements on the same line
        if (
            include_color
            and current.visual.kind in ["vertex", "face"]
            and len(current.visual.vertex_colors)
        ):
            # create a stacked blob with position and color
            v_blob = np.column_stack(
                (current.vertices, to_float(current.visual.vertex_colors[:, :3]))
            )
        else:
            # otherwise just export vertices
            v_blob = current.vertices

            # add the first vertex key and convert the array
        # add the vertices
        export = deque(
            [
                "v "
                + util.array_to_string(
                    v_blob, col_delim=" ", row_delim="\nv ", digits=digits
                )
            ]
        )

        # if include_normals is None then
        # only include if they're already stored
        if include_normals is None:
            include_normals = "vertex_normals" in current._cache.cache

        if include_normals:
            try:
                converted = util.array_to_string(
                    current.vertex_normals,
                    col_delim=" ",
                    row_delim="\nvn ",
                    digits=digits,
                )
                # if vertex normals are stored in cache export them
                face_type.append("vn")
                export.append("vn " + converted)
            except BaseException:
                log.debug("failed to convert vertex normals", exc_info=True)

        # collect materials into a dict
        if include_texture and hasattr(current.visual, "uv"):
            try:
                # get a SimpleMaterial
                material = current.visual.material
                if hasattr(material, "to_simple"):
                    material = material.to_simple()

                # hash the material to avoid duplicates
                hashed = hash(material)
                if hashed not in materials:
                    # get a unique name for the material
                    name = util.unique_name(material.name, materials_name)
                    # add the name to our collection
                    materials_name.add(name)
                    # convert material to an OBJ MTL
                    materials[hashed] = material.to_obj(name=name)

                # get the name of the current material as-stored
                tex_name = materials[hashed][1]

                # export the UV coordinates
                if len(np.shape(getattr(current.visual, "uv", None))) == 2:
                    converted = util.array_to_string(
                        current.visual.uv, col_delim=" ", row_delim="\nvt ", digits=digits
                    )
                    # if vertex texture exists and is the right shape
                    face_type.append("vt")
                    # add the uv coordinates
                    export.append("vt " + converted)
                # add the directive to use the exported material
                export.appendleft(f"usemtl {tex_name}")
            except BaseException:
                log.debug("failed to convert UV coordinates", exc_info=True)

        # the format for a single vertex reference of a face
        face_format = face_formats[tuple(face_type)]
        # add the exported faces to the export if available
        if hasattr(current, "faces"):
            export.append(
                "f "
                + util.array_to_string(
                    current.faces + 1 + counts["v"],
                    col_delim=" ",
                    row_delim="\nf ",
                    value_format=face_format,
                )
            )
        # offset our vertex position
        counts["v"] += len(current.vertices)

        # add object name if found in metadata
        if "name" in current.metadata:
            export.appendleft("\no {}".format(current.metadata["name"]))
        # add this object
        objects.append("\n".join(export))

    # collect files like images to write
    mtl_data = {}
    # combine materials
    if len(materials) > 0:
        # collect text for a single mtllib file
        mtl_lib = []
        # now loop through: keys are garbage hash
        # values are (data, name)
        for data, _ in materials.values():
            for file_name, file_data in data.items():
                if file_name.lower().endswith(".mtl"):
                    # collect mtl lines into single file
                    mtl_lib.append(file_data)
                elif file_name not in mtl_data:
                    # things like images
                    mtl_data[file_name] = file_data
                else:
                    log.warning(f"not writing {file_name}")

        if mtl_name is None:
            # if no name passed set a default
            mtl_name = "material.mtl"

        # prepend a header to the MTL text if requested
        if header is not None:
            prepend = f"# {header}\n\n".encode()
        else:
            prepend = b""

        # save the material data
        mtl_data[mtl_name] = prepend + b"\n\n".join(mtl_lib)
        # add the reference to the MTL file
        objects.appendleft(f"mtllib {mtl_name}")

    if header is not None:
        # add a created-with header to the top of the file
        objects.appendleft(f"# {header}")

    # add a trailing newline
    objects.append("\n")

    # combine elements into a single string
    text = "\n".join(objects)

    # if we have a resolver and have asked to write texture
    if write_texture and resolver is not None and len(materials) > 0:
        # not all resolvers have a write method
        [resolver.write(k, v) for k, v in mtl_data.items()]

    # if we exported texture it changes returned values
    if return_texture:
        return text, mtl_data

    return text


_obj_loaders = {"obj": load_obj}



try:
    from . import generic as g
except BaseException:
    import generic as g


class LightTests(g.unittest.TestCase):
    def test_basic(self):
        for light_class in [
            g.trimesh.scene.lighting.DirectionalLight,
            g.trimesh.scene.lighting.PointLight,
            g.trimesh.scene.lighting.SpotLight,
        ]:
            light = light_class()
            assert isinstance(light.intensity, float)
            assert light.color.shape == (4,)
            assert light.color.dtype == g.np.uint8

    def test_scene(self):
        s = g.get_mesh("duck.dae")
        assert len(s.lights) > 0
        assert isinstance(s.camera, g.trimesh.scene.cameras.Camera)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import numpy as np

from ..util import is_ccw  # NOQA


def concatenate(paths):
    """
    Concatenate multiple paths into a single path.

    Parameters
    -------------
    paths : (n,) Path
      Path objects to concatenate

    Returns
    -------------
    concat : Path, Path2D, or Path3D
      Concatenated result
    """
    # if only one path object just return copy
    if len(paths) == 1:
        return paths[0].copy()

    # upgrade to 3D if we have mixed 2D and 3D paths
    dimensions = {i.vertices.shape[1] for i in paths}
    if len(dimensions) > 1:
        paths = [i.to_3D() if hasattr(i, "to_3D") else i for i in paths]

    # length of vertex arrays
    vert_len = np.array([len(i.vertices) for i in paths])
    # how much to offset each paths vertex indices by
    offsets = np.append(0.0, np.cumsum(vert_len))[:-1].astype(np.int64)

    # resulting entities
    entities = []
    # resulting vertices
    vertices = []
    # resulting metadata
    metadata = {}
    for path, offset in zip(paths, offsets):
        # update metadata
        metadata.update(path.metadata)
        # copy vertices, we will stack later
        vertices.append(path.vertices.copy())
        # copy entity then reindex points
        for entity in path.entities:
            # cleanly copy the entity into a new object
            copied = entity.copy()
            # offset the indexes
            copied.points += offset
            entities.append(copied)
    # generate the single new concatenated path
    # use input types so we don't have circular imports
    concat = type(path)(
        metadata=metadata, entities=entities, vertices=np.vstack(vertices)
    )
    return concat



try:
    from . import generic as g
except BaseException:
    import generic as g


class ReprTest(g.unittest.TestCase):
    def test_repr(self):
        m = g.trimesh.creation.icosphere()
        r = str(m)
        assert "trimesh.Trimesh" in r
        assert "vertices" in r
        assert "faces" in r

        s = m.scene()
        assert isinstance(s, g.trimesh.Scene)
        r = str(s)
        assert "Scene" in r
        assert "geometry" in r

        p = g.trimesh.PointCloud(m.vertices)
        r = str(p)
        assert "trimesh.PointCloud" in r
        assert "vertices" in r

        p = g.trimesh.path.creation.rectangle([[0, 0], [1, 1]])
        assert isinstance(p, g.trimesh.path.Path2D)
        r = str(p)
        assert "trimesh.Path2D" in r
        assert "entities" in r
        assert "vertices" in r

        p = p.to_3D()
        assert isinstance(p, g.trimesh.path.Path3D)
        r = str(p)
        assert "trimesh.Path3D" in r
        assert "entities" in r
        assert "vertices" in r


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
boolean.py
-------------

Do boolean operations on meshes using either Blender or Manifold.
"""

import numpy as np

from . import exceptions, interfaces
from .typed import Callable, Iterable, Optional

try:
    from manifold3d import Manifold, Mesh
except BaseException as E:
    Mesh = exceptions.ExceptionWrapper(E)
    Manifold = exceptions.ExceptionWrapper(E)


def difference(
    meshes: Iterable, engine: Optional[str] = None, check_volume: bool = True, **kwargs
):
    """
    Compute the boolean difference between a mesh an n other meshes.

    Parameters
    ----------
    meshes : sequence of trimesh.Trimesh
      Meshes to be processed.
    engine
      Which backend to use, i.e. 'blender' or 'manifold'
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.

    Returns
    ----------
    difference
      A `Trimesh` that contains `meshes[0] - meshes[1:]`
    """
    if check_volume and not all(m.is_volume for m in meshes):
        raise ValueError("Not all meshes are volumes!")

    return _engines[engine](meshes, operation="difference", **kwargs)


def union(
    meshes: Iterable, engine: Optional[str] = None, check_volume: bool = True, **kwargs
):
    """
    Compute the boolean union between a mesh an n other meshes.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    engine : str
      Which backend to use, i.e. 'blender' or 'manifold'
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.

    Returns
    ----------
    union
      A `Trimesh` that contains the union of all passed meshes.
    """
    if check_volume and not all(m.is_volume for m in meshes):
        raise ValueError("Not all meshes are volumes!")

    result = _engines[engine](meshes, operation="union", **kwargs)
    return result


def intersection(
    meshes: Iterable, engine: Optional[str] = None, check_volume: bool = True, **kwargs
):
    """
    Compute the boolean intersection between a mesh and other meshes.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    engine : str
      Which backend to use, i.e. 'blender' or 'manifold'
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.

    Returns
    ----------
    intersection
      A `Trimesh` that contains the intersection geometry.
    """
    if check_volume and not all(m.is_volume for m in meshes):
        raise ValueError("Not all meshes are volumes!")
    return _engines[engine](meshes, operation="intersection", **kwargs)


def boolean_manifold(
    meshes: Iterable,
    operation: str,
    check_volume: bool = True,
    **kwargs,
):
    """
    Run an operation on a set of meshes using the Manifold engine.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    operation
      Which boolean operation to do.
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.
    kwargs
      Passed through to the `engine`.
    """
    if check_volume and not all(m.is_volume for m in meshes):
        raise ValueError("Not all meshes are volumes!")

    # Convert to manifold meshes
    manifolds = [
        Manifold(
            mesh=Mesh(
                vert_properties=np.array(mesh.vertices, dtype=np.float32),
                tri_verts=np.array(mesh.faces, dtype=np.uint32),
            )
        )
        for mesh in meshes
    ]

    # Perform operations
    if operation == "difference":
        if len(meshes) != 2:
            raise ValueError("Difference only defined over two meshes.")

        result_manifold = manifolds[0] - manifolds[1]
    elif operation == "union":
        result_manifold = reduce_cascade(lambda a, b: a + b, manifolds)
    elif operation == "intersection":
        result_manifold = reduce_cascade(lambda a, b: a ^ b, manifolds)
    else:
        raise ValueError(f"Invalid boolean operation: '{operation}'")

    # Convert back to trimesh meshes
    from . import Trimesh

    result_mesh = result_manifold.to_mesh()
    out_mesh = Trimesh(vertices=result_mesh.vert_properties, faces=result_mesh.tri_verts)

    return out_mesh


def reduce_cascade(operation: Callable, items: Iterable):
    """
    Call a function in a cascaded pairwise way against a
    flat sequence of items. This should produce the same
    result as `functools.reduce` but may be faster for some
    functions that for example perform only as fast as their
    largest input.

    For example on `a b c d e f g` this function would run and return:
        a b
        c d
        e f
        ab cd
        ef g
        abcd efg
     -> abcdefg

    Where `functools.reduce` would run and return:
        a b
        ab c
        abc d
        abcd e
        abcde f
        abcdef g
     -> abcdefg

    Parameters
    ----------
    operation
      The function to call on pairs of items.
    items
      The flat list of items to apply operation against.
    """
    if len(items) == 0:
        return None
    elif len(items) == 2:
        # might as well skip the loop overhead
        return operation(items[0], items[1])

    for _ in range(int(1 + np.log2(len(items)))):
        results = []
        for i in np.arange(len(items) // 2) * 2:
            results.append(operation(items[i], items[i + 1]))

        if len(items) % 2:
            results.append(items[-1])

        items = results

    # logic should have reduced to a single item
    assert len(results) == 1

    return results[0]


# which backend boolean engines
_engines = {
    None: boolean_manifold,
    "manifold": boolean_manifold,
    "blender": interfaces.blender.boolean,
}



"""
sample.py
------------

Randomly sample surface and volume of meshes.
"""

import numpy as np

from . import transformations, util
from .typed import ArrayLike, Integer, NDArray, Number, Optional, float64
from .visual import uv_to_interpolated_color


def sample_surface(
    mesh,
    count: Integer,
    face_weight: Optional[ArrayLike] = None,
    sample_color=False,
    seed=None,
):
    """
    Sample the surface of a mesh, returning the specified
    number of points

    For individual triangle sampling uses this method:
    http://mathworld.wolfram.com/TrianglePointPicking.html

    Parameters
    -----------
    mesh : trimesh.Trimesh
      Geometry to sample the surface of
    count : int
      Number of points to return
    face_weight : None or len(mesh.faces) float
      Weight faces by a factor other than face area.
      If None will be the same as face_weight=mesh.area
    sample_color : bool
      Option to calculate the color of the sampled points.
      Default is False.
    seed : None or int
      If passed as an integer will provide deterministic results
      otherwise pulls the seed from operating system entropy.

    Returns
    ---------
    samples : (count, 3) float
      Points in space on the surface of mesh
    face_index : (count,) int
      Indices of faces for each sampled point
    colors : (count, 4) float
      Colors of each sampled point
      Returns only when the sample_color is True
    """

    if face_weight is None:
        # len(mesh.faces) float, array of the areas
        # of each face of the mesh
        face_weight = mesh.area_faces

    # cumulative sum of weights (len(mesh.faces))
    weight_cum = np.cumsum(face_weight)

    # seed the random number generator as requested
    if seed is None:
        random = np.random.random
    else:
        random = np.random.default_rng(seed).random

    # last value of cumulative sum is total summed weight/area
    face_pick = random(count) * weight_cum[-1]
    # get the index of the selected faces
    face_index = np.searchsorted(weight_cum, face_pick)

    # pull triangles into the form of an origin + 2 vectors
    tri_origins = mesh.vertices[mesh.faces[:, 0]]
    tri_vectors = mesh.vertices[mesh.faces[:, 1:]].copy()
    tri_vectors -= np.tile(tri_origins, (1, 2)).reshape((-1, 2, 3))

    # pull the vectors for the faces we are going to sample from
    tri_origins = tri_origins[face_index]
    tri_vectors = tri_vectors[face_index]

    if sample_color and hasattr(mesh.visual, "uv"):
        uv_origins = mesh.visual.uv[mesh.faces[:, 0]]
        uv_vectors = mesh.visual.uv[mesh.faces[:, 1:]].copy()
        uv_origins_tile = np.tile(uv_origins, (1, 2)).reshape((-1, 2, 2))
        uv_vectors -= uv_origins_tile
        uv_origins = uv_origins[face_index]
        uv_vectors = uv_vectors[face_index]

    # randomly generate two 0-1 scalar components to multiply edge vectors b
    random_lengths = random((len(tri_vectors), 2, 1))

    # points will be distributed on a quadrilateral if we use 2 0-1 samples
    # if the two scalar components sum less than 1.0 the point will be
    # inside the triangle, so we find vectors longer than 1.0 and
    # transform them to be inside the triangle
    random_test = random_lengths.sum(axis=1).reshape(-1) > 1.0
    random_lengths[random_test] -= 1.0
    random_lengths = np.abs(random_lengths)

    # multiply triangle edge vectors by the random lengths and sum
    sample_vector = (tri_vectors * random_lengths).sum(axis=1)

    # finally, offset by the origin to generate
    # (n,3) points in space on the triangle
    samples = sample_vector + tri_origins

    if sample_color:
        if hasattr(mesh.visual, "uv"):
            sample_uv_vector = (uv_vectors * random_lengths).sum(axis=1)
            uv_samples = sample_uv_vector + uv_origins
            texture = mesh.visual.material.image
            colors = uv_to_interpolated_color(uv_samples, texture)
        else:
            colors = mesh.visual.face_colors[face_index]

        return samples, face_index, colors

    return samples, face_index


def volume_mesh(mesh, count: Integer) -> NDArray[float64]:
    """
    Use rejection sampling to produce points randomly
    distributed in the volume of a mesh.


    Parameters
    -----------
    mesh : trimesh.Trimesh
      Geometry to sample
    count : int
      Number of points to return

    Returns
    ---------
    samples : (n, 3) float
      Points in the volume of the mesh where n <= count
    """
    points = (np.random.random((count, 3)) * mesh.extents) + mesh.bounds[0]
    contained = mesh.contains(points)
    samples = points[contained][:count]
    return samples


def volume_rectangular(
    extents, count: Integer, transform: Optional[ArrayLike] = None
) -> NDArray[float64]:
    """
    Return random samples inside a rectangular volume,
    useful for sampling inside oriented bounding boxes.

    Parameters
    -----------
    extents :   (3,) float
      Side lengths of rectangular solid
    count : int
      Number of points to return
    transform : (4, 4) float
      Homogeneous transformation matrix

    Returns
    ---------
    samples : (count, 3) float
      Points in requested volume
    """
    samples = np.random.random((count, 3)) - 0.5
    samples *= extents
    if transform is not None:
        samples = transformations.transform_points(samples, transform)
    return samples


def sample_surface_even(mesh, count: Integer, radius: Optional[Number] = None, seed=None):
    """
    Sample the surface of a mesh, returning samples which are
    VERY approximately evenly spaced. This is accomplished by
    sampling and then rejecting pairs that are too close together.

    Note that since it is using rejection sampling it may return
    fewer points than requested (i.e. n < count). If this is the
    case a log.warning will be emitted.

    Parameters
    -----------
    mesh : trimesh.Trimesh
      Geometry to sample the surface of
    count : int
      Number of points to return
    radius : None or float
      Removes samples below this radius
    seed : None or int
      Provides deterministic values

    Returns
    ---------
    samples : (n, 3) float
      Points in space on the surface of mesh
    face_index : (n,) int
      Indices of faces for each sampled point
    """
    from .points import remove_close

    # guess radius from area
    if radius is None:
        radius = np.sqrt(mesh.area / (3 * count))

    # get points on the surface
    points, index = sample_surface(mesh, count * 3, seed=seed)

    # remove the points closer than radius
    points, mask = remove_close(points, radius)

    # we got all the samples we expect
    if len(points) >= count:
        return points[:count], index[mask][:count]

    # warn if we didn't get all the samples we expect
    util.log.warning(f"only got {len(points)}/{count} samples!")

    return points, index[mask]


def sample_surface_sphere(count: int) -> NDArray[float64]:
    """
    Correctly pick random points on the surface of a unit sphere

    Uses this method:
    http://mathworld.wolfram.com/SpherePointPicking.html

    Parameters
    -----------
    count : int
      Number of points to return

    Returns
    ----------
    points : (count, 3) float
      Random points on the surface of a unit sphere
    """
    # get random values 0.0-1.0
    u, v = np.random.random((2, count))
    # convert to two angles
    theta = np.pi * 2 * u
    phi = np.arccos((2 * v) - 1)
    # convert spherical coordinates to cartesian
    points = util.spherical_to_vector(np.column_stack((theta, phi)))
    return points



try:
    from . import generic as g
except BaseException:
    import generic as g


class ExportTest(g.unittest.TestCase):
    def test_svg(self):
        for d in g.get_2D():
            if g.np.isclose(d.area, 0.0):
                continue
            # export and reload the exported SVG
            loaded = g.trimesh.load(
                g.trimesh.util.wrap_as_stream(d.export(file_type="svg")), file_type="svg"
            )

            # we only have line and arc primitives as SVG
            # export and import
            if all(i.__class__.__name__ in ["Line", "Arc"] for i in d.entities):
                # perimeter should stay the same-ish
                # on export/import
                assert g.np.isclose(d.length, loaded.length, rtol=0.01)

                assert len(d.entities) == len(loaded.entities)

                path_str = g.trimesh.path.exchange.svg_io.export_svg(d, return_path=True)
                assert isinstance(path_str, str)
                assert len(path_str) > 0

    def test_layer(self):
        from shapely.geometry import Point

        # create two disjoint circles and apply layers
        a = g.trimesh.load_path(Point([0, 0]).buffer(1))
        a.apply_layer("ACIRCLE")

        b = g.trimesh.load_path(Point([2, 0]).buffer(1))
        b.apply_layer("BCIRCLE")

        assert id(a.entities[0]._metadata) != id(b.entities[0]._metadata)

        # combine two circles
        c = a + b

        # should be combined
        assert g.np.isclose(c.area, a.area + b.area)

        # export C with just layer of A
        aX = g.trimesh.load(
            g.io_wrap(c.export(file_type="svg", only_layers=["ACIRCLE"])), file_type="svg"
        )

        # export C with all layers
        cX = g.trimesh.load(
            g.io_wrap(c.export(file_type="svg", only_layers=None)), file_type="svg"
        )

        assert len(cX.entities) == len(c.entities)
        # should have skipped the layers
        assert len(aX.entities) == 1

        # make
        aR = g.trimesh.load(
            g.io_wrap(c.export(file_type="dxf", only_layers=["ACIRCLE"])), file_type="dxf"
        )

        assert g.np.isclose(aR.area, a.area)

    def test_trans(self):
        from trimesh.path.exchange.svg_io import transform_to_matrices as tf

        # empty strings shouldn't have matrix
        assert len(tf("")) == 0

        # check translate with different whitespace
        a = tf("translate(1.1, 2.2      )")
        assert len(a) == 1
        assert g.np.allclose(a[0], [[1, 0, 1.1], [0, 1, 2.2], [0, 0, 1]])
        a = tf(" translate(1.1    1.2   )       ")
        assert len(a) == 1
        assert g.np.allclose(a[0], [[1, 0, 1.1], [0, 1, 1.2], [0, 0, 1]])

        a = tf(
            " translate(1.1    1.2   )       "
            + "matrix (  {} {} {} {} {} {})".format(*g.np.arange(6))
        )
        assert len(a) == 2
        # check the translate
        assert g.np.allclose(a[0], [[1, 0, 1.1], [0, 1, 1.2], [0, 0, 1]])
        # check the matrix string
        assert g.np.allclose(a[1], [[0, 2, 4], [1, 3, 5], [0, 0, 1]])

    def test_roundtrip(self):
        """
        Check to make sure a roundtrip from both a Scene and a
        Path2D results in the same file on both sides
        """
        for fn in ["2D/250_cycloidal.DXF", "2D/tray-easy1.dxf"]:
            p = g.get_mesh(fn)
            assert isinstance(p, g.trimesh.path.Path2D)
            # load the exported SVG
            r = g.trimesh.load(
                g.trimesh.util.wrap_as_stream(p.export(file_type="svg")), file_type="svg"
            )
            assert isinstance(r, g.trimesh.path.Path2D)
            assert g.np.isclose(r.length, p.length)
            assert g.np.isclose(r.area, p.area)

            assert set(r.metadata.keys()) == set(p.metadata.keys())

            s = g.trimesh.scene.split_scene(p)
            as_svg = s.export(file_type="svg")
            assert isinstance(s, g.trimesh.Scene)
            r = g.trimesh.load(g.trimesh.util.wrap_as_stream(as_svg), file_type="svg")
            assert isinstance(r, g.trimesh.Scene)
            assert s.metadata == r.metadata

            # make sure every geometry name matches exactly
            assert set(s.geometry.keys()) == set(r.geometry.keys())

            # check to see if every geometry has the same metadata
            for geom in s.geometry.keys():
                a, b = s.geometry[geom], r.geometry[geom]
                assert a.metadata == b.metadata
                assert g.np.isclose(a.length, b.length)
                assert g.np.isclose(a.area, b.area)
                assert a.body_count == b.body_count

            assert r.metadata["file_path"].endswith(fn[3:])


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
repair.py
-------------

Fill holes and fix winding and normals of meshes.
"""

import numpy as np

from . import graph, triangles
from .constants import log
from .geometry import faces_to_edges
from .grouping import group_rows

try:
    import networkx as nx
except BaseException as E:
    # create a dummy module which will raise the ImportError
    # or other exception only when someone tries to use networkx
    from .exceptions import ExceptionWrapper

    nx = ExceptionWrapper(E)

try:
    from .path.exchange.misc import faces_to_path
except BaseException as E:
    from .exceptions import ExceptionWrapper

    faces_to_path = ExceptionWrapper(E)


def fix_winding(mesh):
    """
    Traverse and change mesh faces in-place to make sure
    winding is correct with edges on adjacent faces in
    opposite directions.

    Parameters
    -------------
    mesh : Trimesh
      Source geometry to alter in-place.
    """
    # anything we would fix is already done
    if mesh.is_winding_consistent:
        return

    graph_all = nx.from_edgelist(mesh.face_adjacency)
    flipped = 0

    faces = mesh.faces.view(np.ndarray).copy()

    # we are going to traverse the graph using BFS
    # start a traversal for every connected component
    for components in nx.connected_components(graph_all):
        # get a subgraph for this component
        g = graph_all.subgraph(components)
        # get the first node in the graph in a way that works on nx's
        # new API and their old API
        start = next(iter(g.nodes()))

        # we traverse every pair of faces in the graph
        # we modify mesh.faces and mesh.face_normals in place
        for face_pair in nx.bfs_edges(g, start):
            # for each pair of faces, we convert them into edges,
            # find the edge that both faces share and then see if edges
            # are reversed in order as you would expect
            # (2, ) int
            face_pair = np.ravel(face_pair)
            # (2, 3) int
            pair = faces[face_pair]
            # (6, 2) int
            edges = faces_to_edges(pair)
            overlap = group_rows(np.sort(edges, axis=1), require_count=2)
            if len(overlap) == 0:
                # only happens on non-watertight meshes
                continue
            edge_pair = edges[overlap[0]]
            if edge_pair[0][0] == edge_pair[1][0]:
                # if the edges aren't reversed, invert the order of one face
                flipped += 1
                faces[face_pair[1]] = faces[face_pair[1]][::-1]

    if flipped > 0:
        mesh.faces = faces

    log.debug("flipped %d/%d edges", flipped, len(mesh.faces) * 3)


def fix_inversion(mesh, multibody=False):
    """
    Check to see if a mesh has normals pointing "out."

    Parameters
    -------------
    mesh : trimesh.Trimesh
      Mesh to fix in-place.
    multibody : bool
      If True will try to fix normals on every body
    """
    if not mesh.is_watertight:
        # this will make things worse for non-watertight meshes
        return

    if multibody:
        groups = graph.connected_components(mesh.face_adjacency)
        # escape early for single body
        if len(groups) == 1:
            if mesh.volume < 0.0:
                mesh.invert()
            return
        # mask of faces to flip
        flip = np.zeros(len(mesh.faces), dtype=bool)
        # save these to avoid thrashing cache
        tri = mesh.triangles
        cross = mesh.triangles_cross
        # indexes of mesh.faces, not actual faces
        for faces in groups:
            # calculate the volume of the submesh faces
            volume = triangles.mass_properties(
                tri[faces], crosses=cross[faces], skip_inertia=True
            )["volume"]
            # if that volume is negative it is either
            # inverted or just total garbage
            if volume < 0.0:
                flip[faces] = True
        # one or more faces needs flipping
        if flip.any():
            # flip normals of necessary faces
            if "face_normals" in mesh._cache:
                normals = mesh.face_normals.copy()
                normals[flip] *= -1.0
            else:
                normals = None
            # flip faces
            mesh.faces[flip] = np.fliplr(mesh.faces[flip])
            if normals is not None:
                mesh.face_normals = normals

    elif mesh.volume < 0.0:
        mesh.invert()


def fix_normals(mesh, multibody=False):
    """
    Fix the winding and direction of a mesh face and
    face normals in-place.

    Really only meaningful on watertight meshes but will orient all
    faces and winding in a uniform way for non-watertight face
    patches as well.

    Parameters
    -------------
    mesh : trimesh.Trimesh
      Mesh to fix normals on
    multibody : bool
      if True try to correct normals direction
      on every body rather than just one

    Notes
    --------------
    mesh.faces : will flip columns on inverted faces
    """
    # traverse face adjacency to correct winding
    fix_winding(mesh)
    # check to see if a mesh is inverted
    fix_inversion(mesh, multibody=multibody)


def broken_faces(mesh, color=None):
    """
    Return the index of faces in the mesh which break the
    watertight status of the mesh.

    Parameters
    --------------
    mesh : trimesh.Trimesh
      Mesh to check broken faces on
    color: (4,) uint8 or None
      Will set broken faces to this color if not None

    Returns
    ---------------
    broken : (n, ) int
      Indexes of mesh.faces
    """
    adjacency = nx.from_edgelist(mesh.face_adjacency)
    broken = [k for k, v in dict(adjacency.degree()).items() if v != 3]
    broken = np.array(broken)
    if color is not None and broken.size != 0:
        # if someone passed a broken color
        color = np.array(color)
        if not (color.shape == (4,) or color.shape == (3,)):
            color = [255, 0, 0, 255]
        mesh.visual.face_colors[broken] = color
    return broken


def fill_holes(mesh):
    """
    Fill single- triangle holes on triangular meshes by adding
    new triangles to fill the holes. New triangles will have
    proper winding and normals, and if face colors exist the color
    of the last face will be assigned to the new triangles.

    Parameters
    ------------
    mesh : trimesh.Trimesh
      Mesh will be repaired in- place
    """

    def hole_to_faces(hole):
        """
        Given a loop of vertex indices  representing a hole
        turn it into triangular faces.
        If unable to do so, return None

        Parameters
        -----------
        hole : (n,) int
          Ordered loop of vertex indices

        Returns
        ---------
        faces : (n, 3) int
          New faces
        vertices : (m, 3) float
          New vertices
        """
        hole = np.asanyarray(hole)
        # the case where the hole is just a single missing triangle
        if len(hole) == 3:
            return [hole], []
        # the hole is a quad, which we fill with two triangles
        if len(hole) == 4:
            face_A = hole[[0, 1, 2]]
            face_B = hole[[2, 3, 0]]
            return [face_A, face_B], []
        return [], []

    if len(mesh.faces) < 3:
        return False

    if mesh.is_watertight:
        return True

    # we know that in a watertight mesh every edge will be included twice
    # thus every edge which appears only once is part of a hole boundary
    boundary_groups = group_rows(mesh.edges_sorted, require_count=1)

    # mesh is not watertight and we have too few edges
    # edges to do a repair
    # since we haven't changed anything return False
    if len(boundary_groups) < 3:
        return False

    boundary_edges = mesh.edges[boundary_groups]
    index_as_dict = [{"index": i} for i in boundary_groups]

    # we create a graph of the boundary edges, and find cycles.
    g = nx.from_edgelist(np.column_stack((boundary_edges, index_as_dict)))
    new_faces = []
    new_vertex = []
    for hole in nx.cycle_basis(g):
        # convert the hole, which is a polygon of vertex indices
        # to triangles and new vertices
        faces, vertex = hole_to_faces(hole=hole)
        if len(faces) == 0:
            continue
        # remeshing returns new vertices as negative indices, so change those
        # to absolute indices which won't be screwed up by the later appends
        faces = np.array(faces)
        faces[faces < 0] += len(new_vertex) + len(mesh.vertices) + len(vertex)
        new_vertex.extend(vertex)
        new_faces.extend(faces)
    new_faces = np.array(new_faces)
    new_vertex = np.array(new_vertex)

    if len(new_faces) == 0:
        # no new faces have been added, so nothing further to do
        # the mesh is NOT watertight, as boundary groups exist
        # but we didn't add any new faces to fill them in
        return False

    for face_index, face in enumerate(new_faces):
        # we compare the edge from the new face with
        # the boundary edge from the source mesh
        edge_test = face[:2]
        edge_boundary = mesh.edges[g.get_edge_data(*edge_test)["index"]]

        # in a well constructed mesh, the winding is such that adjacent triangles
        # have reversed edges to each other. Here we check to make sure the
        # edges are reversed, and if they aren't we simply reverse the face
        reversed = edge_test[0] == edge_boundary[1]
        if not reversed:
            new_faces[face_index] = face[::-1]

    # stack vertices into clean (n, 3) float
    if len(new_vertex) != 0:
        new_vertices = np.vstack((mesh.vertices, new_vertex))
    else:
        new_vertices = mesh.vertices

    # try to save face normals if we can
    if "face_normals" in mesh._cache.cache:
        cached_normals = mesh._cache.cache["face_normals"]
    else:
        cached_normals = None

    # also we can remove any zero are triangles by masking here
    new_normals, valid = triangles.normals(new_vertices[new_faces])
    # all the added faces were broken
    if not valid.any():
        return False

    # this is usually the case where two vertices of a triangle are just
    # over tol.merge apart, but the normal calculation is screwed up
    # these could be fixed by merging the vertices in question here:
    # if not valid.all():
    if mesh.visual.defined and mesh.visual.kind == "face":
        color = mesh.visual.face_colors
    else:
        color = None

    # apply the new faces and vertices
    mesh.faces = np.vstack((mesh._data["faces"], new_faces[valid]))
    mesh.vertices = new_vertices

    # dump the cache and set id to the new hash
    mesh._cache.verify()

    # save us a normals recompute if we can
    if cached_normals is not None:
        mesh.face_normals = np.vstack((cached_normals, new_normals))

    # this is usually the case where two vertices of a triangle are just
    # over tol.merge apart, but the normal calculation is screwed up
    # these could be fixed by merging the vertices in question here:
    # if not valid.all():
    if color is not None:
        # if face colors exist, assign the last face color to the new faces
        # note that this is a little cheesey, but it is very inexpensive and
        # is the right thing to do if the mesh is a single color.
        color_shape = np.shape(color)
        if len(color_shape) == 2:
            new_colors = np.tile(color[-1], (np.sum(valid), 1))
            new_colors = np.vstack((color, new_colors))
            mesh.visual.face_colors = new_colors

    log.debug("Filled in mesh with %i triangles", np.sum(valid))
    return mesh.is_watertight


def stitch(mesh, faces=None, insert_vertices=False):
    """
    Create a fan stitch over the boundary of the specified
    faces. If the boundary is non-convex a triangle fan
    is going to be extremely wonky.

    Parameters
    -----------
    vertices : (n, 3) float
      Vertices in space.
    faces : (n,) int
      Face indexes to stitch with triangle fans.
    insert_vertices : bool
      Allow stitching to insert new vertices?

    Returns
    ----------
    fan : (m, 3) int
      New triangles referencing mesh.vertices.
    vertices : (p, 3) float
      Inserted vertices (only returned `if insert_vertices`)
    """
    if faces is None:
        faces = np.arange(len(mesh.faces))

    # get a sequence of vertex indices representing the
    # boundary of the specified faces
    # will be referencing the same indexes of `mesh.vertices`
    points = [
        e.points
        for e in faces_to_path(mesh, faces)["entities"]
        if len(e.points) > 3 and e.points[0] == e.points[-1]
    ]

    # get properties to avoid querying in loop
    vertices = mesh.vertices
    normals = mesh.face_normals

    # find which faces are associated with an edge
    edges_face = mesh.edges_face
    tree_edge = mesh.edges_sorted_tree

    if insert_vertices:
        # create one new vertex per curve at the centroid
        centroids = np.array([vertices[p].mean(axis=0) for p in points])
        # save the original length of the vertices
        count = len(vertices)
        # for the normal check stack our local vertices
        vertices = np.vstack((vertices, centroids))
        # create a triangle between our new centroid vertex
        # and each one of the boundary curves
        fan = [
            np.column_stack((np.ones(len(p) - 1, dtype=int) * (count + i), p[:-1], p[1:]))
            for i, p in enumerate(points)
        ]
    else:
        # since we're not allowed to insert new vertices
        # create a triangle fan for each boundary curve
        fan = [
            np.column_stack((np.ones(len(p) - 3, dtype=int) * p[0], p[1:-2], p[2:-1]))
            for p in points
        ]

    # now we do a normal check against an adjacent face
    # to see if each region needs to be flipped
    for i, t in zip(range(len(fan)), fan):
        # get the edges from the original mesh
        # for the first `n` new triangles
        e = t[:10, 1:].copy()
        e.sort(axis=1)

        # find which indexes of `mesh.edges` these
        # new edges correspond with by finding edges
        # that exactly correspond with the tree
        query = tree_edge.query_ball_point(e, r=1e-10)
        if len(query) == 0:
            continue
        # stack all the indices that exist
        edge_index = np.concatenate(query)

        # get the normals from the original mesh
        original = normals[edges_face[edge_index]]

        # calculate the normals for a few new faces
        check, valid = triangles.normals(vertices[t[:3]])
        if not valid.any():
            continue
        # take the first valid normal from our new faces
        check = check[0]

        # if our new faces are reversed from the original
        # Adjacent face flip them along their axis
        sign = np.dot(original, check)
        if sign.mean() < 0:
            fan[i] = np.fliplr(t)

    fan = np.vstack(fan)

    if insert_vertices:
        return fan, centroids
    return fan



"""
permutate.py
-------------

Randomly deform meshes in different ways.
"""

import numpy as np

from . import transformations, util
from . import triangles as triangles_module


def transform(mesh, translation_scale=1000.0):
    """
    Return a permutated variant of a mesh by randomly reordering faces
    and rotatating + translating a mesh by a random matrix.

    Parameters
    ----------
    mesh : trimesh.Trimesh
      Mesh, will not be altered by this function

    Returns
    ----------
    permutated : trimesh.Trimesh
      Mesh with same faces as input mesh but reordered
      and rigidly transformed in space.
    """
    # rotate and translate randomly
    matrix = transformations.random_rotation_matrix(translate=translation_scale)

    # randomly re-order triangles
    triangles = np.random.permutation(mesh.triangles).reshape((-1, 3))
    # apply rigid transform
    triangles = transformations.transform_points(triangles, matrix)

    # extract the class from the input object
    mesh_type = util.type_named(mesh, "Trimesh")
    # generate a new mesh from the permutated data
    permutated = mesh_type(**triangles_module.to_kwargs(triangles.reshape((-1, 3, 3))))

    return permutated


def noise(mesh, magnitude=None):
    """
    Add gaussian noise to every vertex of a mesh, making
    no effort to maintain topology or sanity.

    Parameters
    ----------
    mesh : trimesh.Trimesh
      Input geometry, will not be altered
    magnitude : float
      What is the maximum distance per axis we can displace a vertex.
      If None, value defaults to (mesh.scale / 100.0)

    Returns
    ----------
    permutated : trimesh.Trimesh
      Input mesh with noise applied
    """
    if magnitude is None:
        magnitude = mesh.scale / 100.0

    random = (np.random.random(mesh.vertices.shape) - 0.5) * magnitude
    vertices_noise = mesh.vertices.copy() + random

    # make sure we've re- ordered faces randomly
    triangles = np.random.permutation(vertices_noise[mesh.faces])

    mesh_type = util.type_named(mesh, "Trimesh")
    permutated = mesh_type(**triangles_module.to_kwargs(triangles))

    return permutated


def tessellation(mesh):
    """
    Subdivide each face of a mesh into three faces with the new vertex
    randomly placed inside the old face.

    This produces a mesh with exactly the same surface area and volume
    but with different tessellation.

    Parameters
    ------------
    mesh : trimesh.Trimesh
      Input geometry

    Returns
    ----------
    permutated : trimesh.Trimesh
      Mesh with remeshed facets
    """
    # create random barycentric coordinates for each face
    # pad all coordinates by a small amount to bias new vertex towards center
    barycentric = np.random.random(mesh.faces.shape) + 0.05
    barycentric /= barycentric.sum(axis=1).reshape((-1, 1))

    # create one new vertex somewhere in a face
    vertex_face = (barycentric.reshape((-1, 3, 1)) * mesh.triangles).sum(axis=1)
    vertex_face_id = np.arange(len(vertex_face)) + len(mesh.vertices)

    # new vertices are the old vertices stacked on the vertices in the faces
    vertices = np.vstack((mesh.vertices, vertex_face))
    # there are three new faces per old face, and we maintain correct winding
    faces = np.vstack(
        (
            np.column_stack((mesh.faces[:, [0, 1]], vertex_face_id)),
            np.column_stack((mesh.faces[:, [1, 2]], vertex_face_id)),
            np.column_stack((mesh.faces[:, [2, 0]], vertex_face_id)),
        )
    )
    # make sure the order of the faces is permutated
    faces = np.random.permutation(faces)

    mesh_type = util.type_named(mesh, "Trimesh")
    permutated = mesh_type(vertices=vertices, faces=faces)
    return permutated


class Permutator:
    def __init__(self, mesh):
        """
        A convenience object to get permutated versions of a mesh.
        """
        self._mesh = mesh

    def transform(self, translation_scale=1000):
        return transform(self._mesh, translation_scale=translation_scale)

    def noise(self, magnitude=None):
        return noise(self._mesh, magnitude)

    def tessellation(self):
        return tessellation(self._mesh)


try:
    # copy the function docstrings to the helper object
    Permutator.noise.__doc__ = noise.__doc__
    Permutator.transform.__doc__ = transform.__doc__
    Permutator.tessellation.__doc__ = tessellation.__doc__
except AttributeError:
    # no docstrings in Python2
    pass



try:
    from . import generic as g
except BaseException:
    import generic as g

tol_norm = 1e-10


class AlignTests(g.unittest.TestCase):
    def test_align(self):
        """
        Test aligning two 3D vectors
        """

        # function we're testing
        align = g.trimesh.geometry.align_vectors
        is_rigid = g.trimesh.transformations.is_rigid

        # start with some edge cases and make sure the transform works
        target = g.np.array([0, 0, -1], dtype=g.np.float64)
        vectors = g.np.vstack(
            (
                g.trimesh.unitize(g.random((1000, 3)) - 0.5),
                g.random((1000, 3)) - 0.5,
                [-target, target],
                g.trimesh.util.generate_basis(target),
                [
                    [7.12106798e-07, -7.43194705e-08, 1.00000000e00],
                    [0, 0, -1],
                    [1e-4, 1e-4, -1],
                ],
            )
        )

        # collect errors
        norms = []
        unitized = g.trimesh.unitize(vectors)
        for unit_dest, dest in zip(unitized[-10:], vectors[-10:]):
            for unit, vector in zip(unitized, vectors):
                T, _a = align(vector, dest, return_angle=True)
                assert is_rigid(T)
                assert g.np.isclose(g.np.linalg.det(T), 1.0)
                # rotate vector with transform
                check = g.np.dot(T[:3, :3], unit)
                # compare to target vector
                norm = g.np.linalg.norm(check - unit_dest)
                norms.append(norm)
                assert norm < tol_norm

        norms = g.np.array(norms)
        # these vectors should be perpendicular and zero
        angles = [
            align(i, target, return_angle=True)[1]
            for i in g.trimesh.util.generate_basis(target)
        ]
        assert g.np.allclose(angles, [g.np.pi / 2, g.np.pi / 2, 0.0])

    def test_range(self):
        # function we're testing
        align = g.trimesh.geometry.align_vectors
        is_rigid = g.trimesh.transformations.is_rigid

        # generate angles from 0 to 180 degrees
        angles = g.np.linspace(0.0, g.np.pi / 1e7, 10000)
        # generate on- plane vectors
        vectors = g.np.column_stack(
            (g.np.cos(angles), g.np.sin(angles), g.np.zeros(len(angles)))
        )

        # rotate them arbitrarily off the plane just for funsies
        vectors = g.trimesh.transform_points(vectors, g.transforms[20])

        for angle, vector in zip(angles, vectors):
            g.trimesh.util.generate_basis(vector)
            # check alignment to first vector
            # which was created with zero angle
            T, a = align(vector, vectors[0], return_angle=True)
            assert is_rigid(T)
            # check to make sure returned angle corresponds with truth

            assert g.np.isclose(a, angle, atol=1e-6)

            # check to make sure returned transform is correct
            check = g.np.dot(T[:3, :3], vector)
            norm = g.np.linalg.norm(check - vectors[0])

            assert norm < tol_norm

    def test_rigid(self):
        # check issues with near-reversed vectors not returning rigid
        align = g.trimesh.geometry.align_vectors
        T = align([0, 0, -1], [-1e-17, 1e-17, 1])
        assert g.np.isclose(g.np.linalg.det(T), 1.0)

        T = align([0, 0, -1], [-1e-4, 1e-4, 1])
        assert g.np.isclose(g.np.linalg.det(T), 1.0)

        vector_1 = g.np.array([7.12106798e-07, -7.43194705e-08, 1.00000000e00])
        vector_2 = g.np.array([0, 0, -1])
        T, _angle = align(vector_1, vector_2, return_angle=True)
        assert g.np.isclose(g.np.linalg.det(T), 1.0)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class MFTest(g.unittest.TestCase):
    def test_3MF(self):
        # an assembly with instancing
        s = g.get_mesh("counterXP.3MF")

        # should be 2 unique meshes
        assert len(s.geometry) == 2
        # should be 6 instances around the scene
        assert len(s.graph.nodes_geometry) == 6
        assert all(m.is_volume for m in s.geometry.values())

        # a single body 3MF assembly
        s = g.get_mesh("featuretype.3MF")
        # should be 2 unique meshes
        assert len(s.geometry) == 1
        # should be 6 instances around the scene
        assert len(s.graph.nodes_geometry) == 1

    def test_units(self):
        # test our unit conversion function
        converter = g.trimesh.units.unit_conversion
        # these are the units listed in the 3MF spec as valid
        units = ["micron", "millimeter", "centimeter", "inch", "foot", "meter"]
        # check conversion factor for all valid 3MF units
        assert all(converter(u, "inches") > 1e-12 for u in units)

    def test_kwargs(self):
        # check if kwargs are properly passed to geometries
        s = g.get_mesh("P_XPM_0331_01.3mf")
        assert all(len(v.vertices) == 4 for v in s.geometry.values())

        s = g.get_mesh("P_XPM_0331_01.3mf", process=False)
        assert all(len(v.vertices) == 5 for v in s.geometry.values())

    def test_names(self):
        # check if two different objects with the same name are correctly
        # processed
        s = g.get_mesh("cube_and_sphere_same_name.3mf")
        assert len(s.geometry) == 2

    def test_roundtrip(self):
        if g.sys.version_info < (3, 6):
            g.log.warning("relies on > Python 3.5")
            return

        # test a scene round-tripped through the
        # 3MF exporter and importer
        s = g.get_mesh("cycloidal.3DXML")
        assert len(s.geometry) == 13

        # export and reload
        r = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(s.export(file_type="3mf")),
            file_type="3mf",
        )

        assert set(s.geometry.keys()) == set(r.geometry.keys())
        assert g.np.allclose(s.bounds, r.bounds)
        assert g.np.isclose(s.area, r.area, rtol=1e-3)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import json
import os
import warnings

from ..typed import Dict, Stream
from ..util import decode_text, wrap_as_stream

# find the current absolute path to this directory
_pwd = os.path.expanduser(os.path.abspath(os.path.dirname(__file__)))

# once resources are loaded cache them
_cache = {}


def get(
    name: str, decode: bool = True, decode_json: bool = False, as_stream: bool = False
):
    """
    DERECATED JANUARY 2025 REPLACE WITH TYPED `get_json`, `get_string`, etc.
    """
    warnings.warn(
        "`trimesh.resources.get` is deprecated "
        + "and will be removed in January 2025: "
        + "replace with typed `trimesh.resources.get_*type*`",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return _get(name=name, decode=decode, decode_json=decode_json, as_stream=as_stream)


def _get(name: str, decode: bool, decode_json: bool, as_stream: bool):
    """
    Get a resource from the `trimesh/resources` folder.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`
    decode : bool
      Whether or not to decode result as UTF-8
    decode_json : bool
      Run `json.loads` on resource if True.
    as_stream : bool
      Return as a file-like object

    Returns
    -------------
    resource : str, bytes, or decoded JSON
      File data
    """
    # key by name and decode
    cache_key = (name, bool(decode), bool(decode_json), bool(as_stream))
    cached = _cache.get(cache_key)
    if hasattr(cached, "seek"):
        cached.seek(0)
    if cached is not None:
        return cached

    # get the resource using relative names
    with open(os.path.join(_pwd, name), "rb") as f:
        resource = f.read()

    # make sure we return it as a string if asked
    if decode:
        # will decode into text if possibly
        resource = decode_text(resource)

    if decode_json:
        resource = json.loads(resource)
    elif as_stream:
        resource = wrap_as_stream(resource)

    # store for later access
    _cache[cache_key] = resource

    return resource


def get_schema(name: str) -> Dict:
    """
    Load a schema and evaluate the referenced files.

    Parameters
    ------------
    name : str
      Filename of schema.

    Returns
    ----------
    schema
      Loaded and resolved schema.
    """
    from ..resolvers import FilePathResolver
    from ..schemas import resolve

    # get a resolver for our base path
    resolver = FilePathResolver(os.path.join(_pwd, "schema", name))
    # recursively load $ref keys
    return resolve(json.loads(decode_text(resolver.get(name))), resolver=resolver)


def get_json(name: str) -> Dict:
    """
    Get a resource from the `trimesh/resources` folder as a decoded string.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`

    Returns
    -------------
    resource
      File data decoded from JSON.
    """
    return _get(name, decode=True, decode_json=True, as_stream=False)


def get_string(name: str) -> str:
    """
    Get a resource from the `trimesh/resources` folder as a decoded string.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`

    Returns
    -------------
    resource
      File data as a string.
    """
    return _get(name, decode=True, decode_json=False, as_stream=False)


def get_bytes(name: str) -> bytes:
    """
    Get a resource from the `trimesh/resources` folder as binary data.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`

    Returns
    -------------
    resource
      File data as raw bytes.
    """
    return _get(name, decode=False, decode_json=False, as_stream=False)


def get_stream(name: str) -> Stream:
    """
    Get a resource from the `trimesh/resources` folder as a binary stream.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`

    Returns
    -------------
    resource
      File data as a binary stream.
    """

    return _get(name, decode=False, decode_json=False, as_stream=True)



"""
notebook.py
-------------

Render trimesh.Scene objects in HTML
and jupyter notebooks using three.js
"""

import base64
import os

# for our template
from .. import resources, util


def scene_to_html(scene):
    """
    Return HTML that will render the scene using
    GLTF/GLB encoded to base64 loaded by three.js

    Parameters
    --------------
    scene : trimesh.Scene
      Source geometry

    Returns
    --------------
    html : str
      HTML containing embedded geometry
    """
    # fetch HTML template from ZIP archive
    # it is bundling all of three.js so compression is nice
    base = (
        util.decompress(resources.get_bytes("templates/viewer.zip"), file_type="zip")[
            "viewer.html.template"
        ]
        .read()
        .decode("utf-8")
    )
    # make sure scene has camera populated before export
    _ = scene.camera
    # get export as bytes
    data = scene.export(file_type="glb")
    # encode as base64 string
    encoded = base64.b64encode(data).decode("utf-8")
    # replace keyword with our scene data
    result = base.replace("$B64GLTF", encoded)

    return result


def scene_to_notebook(scene, height=500, **kwargs):
    """
    Convert a scene to HTML containing embedded geometry
    and a three.js viewer that will display nicely in
    an IPython/Jupyter notebook.

    Parameters
    -------------
    scene : trimesh.Scene
      Source geometry

    Returns
    -------------
    html : IPython.display.HTML
      Object containing rendered scene
    """
    # keep as soft dependency
    from IPython import display

    # convert scene to a full HTML page
    as_html = scene_to_html(scene=scene)

    # escape the quotes in the HTML
    srcdoc = as_html.replace('"', "&quot;")
    # embed this puppy as the srcdoc attr of an IFframe
    # I tried this a dozen ways and this is the only one that works
    # display.IFrame/display.Javascript really, really don't work
    # div is to avoid IPython's pointless hardcoded warning
    embedded = display.HTML(
        " ".join(
            [
                '<div><iframe srcdoc="{srcdoc}"',
                'width="100%" height="{height}px"',
                'style="border:none;"></iframe></div>',
            ]
        ).format(srcdoc=srcdoc, height=height)
    )
    return embedded


def in_notebook():
    """
    Check to see if we are in an IPython or Jypyter notebook.

    Returns
    -----------
    in_notebook : bool
      Returns True if we are in a notebook
    """
    try:
        # function returns IPython context, but only in IPython
        ipy = get_ipython()  # NOQA
        # we only want to render rich output in notebooks
        # in terminals we definitely do not want to output HTML
        name = str(ipy.__class__).lower()
        terminal = "terminal" in name

        # spyder uses ZMQshell, and can appear to be a notebook
        spyder = "_" in os.environ and "spyder" in os.environ["_"]

        # assume we are in a notebook if we are not in
        # a terminal and we haven't been run by spyder
        notebook = (not terminal) and (not spyder)

        return notebook

    except BaseException:
        return False



try:
    from . import generic as g
except BaseException:
    import generic as g


class VisualTest(g.unittest.TestCase):
    def test_visual(self):
        mesh = g.get_mesh("featuretype.STL")

        # stl shouldn't have any visual properties defined
        assert not mesh.visual.defined

        for facet in mesh.facets:
            mesh.visual.face_colors[facet] = g.trimesh.visual.random_color()

        assert mesh.visual.defined
        assert not mesh.visual.transparency

        mesh.visual.face_colors[0] = [10, 10, 10, 130]
        assert mesh.visual.transparency

    def test_concatenate(self):
        a = g.get_mesh("ballA.off")
        b = g.get_mesh("ballB.off")

        a.visual.face_colors = [255, 0, 0]
        r = a + b
        assert any(g.np.ptp(r.visual.face_colors, axis=0) > 1)

    def test_concatenate_empty_mesh(self):
        box = g.get_mesh("box.STL")

        mesh_fcolor = box.copy()
        mesh_fcolor.visual.face_colors = [255, 0, 0]

        mesh_vcolor = box.copy()
        mesh_vcolor.visual.vertex_colors = [0, 0, 255]

        mesh_empty = g.trimesh.Trimesh()

        r_left_fcolor = mesh_fcolor + mesh_empty
        r_right_fcolor = mesh_empty + mesh_fcolor
        r_left_vcolor = mesh_vcolor + mesh_empty
        r_right_vcolor = mesh_empty + mesh_vcolor
        r_empty = mesh_empty + mesh_empty

        for visual_face in [r_left_fcolor.visual, r_right_fcolor.visual]:
            assert (visual_face.face_colors == mesh_fcolor.visual.face_colors).all()
            assert visual_face.kind == "face"

        for visual_vert in [r_left_vcolor.visual, r_right_vcolor.visual]:
            assert (visual_vert.vertex_colors == mesh_vcolor.visual.vertex_colors).all()
            assert visual_vert.kind == "vertex"

        assert len(r_empty.visual.face_colors) == 0
        assert r_empty.visual.kind is None

    def test_data_model(self):
        """
        Test the probably too- magical color caching and storage
        system.
        """
        m = g.get_mesh("featuretype.STL")
        test_color = [255, 0, 0, 255]
        test_color_2 = [0, 255, 0, 255]
        test_color_transparent = [25, 33, 0, 146]

        # there should be nothing in the cache or DataStore when
        # starting
        assert len(m.visual._cache) == 0
        assert len(m.visual._data) == 0
        # no visuals have been defined so this should be None
        assert m.visual.kind is None
        assert not m.visual.defined

        # this should cause colors to be generated into cache
        initial_id = id(m.visual.face_colors)
        assert m.visual.face_colors.shape[0] == len(m.faces)
        assert id(m.visual.face_colors) == initial_id
        # the values should be in the cache and not in data
        assert len(m.visual._cache) > 0
        assert len(m.visual._data) == 0
        assert not m.visual.defined
        assert not m.visual.transparency

        # this should move the color from cache to data
        m.visual.face_colors[0] = test_color
        # the operation should have moved the colors into data but
        # the object ID should be the same as on creation
        # assert id(m.visual.face_colors) == initial_id
        # the color assignment inside the array should have worked
        assert (m.visual.face_colors[0] == test_color).all()
        # the rest of the colors should be unchanged
        assert (m.visual.face_colors[1] != test_color).any()
        assert len(m.visual._data) >= 1
        assert m.visual.kind == "face"
        assert m.visual.defined
        assert not m.visual.transparency

        # set all face colors to test color
        m.visual.face_colors = test_color
        assert (m.visual.face_colors == test_color).all()
        # assert len(m.visual._cache) == 0
        # should be just material and face information
        assert len(m.visual._data.data) >= 1
        assert m.visual.kind == "face"
        assert bool((m.visual.vertex_colors == test_color).all())
        assert m.visual.defined
        assert not m.visual.transparency

        # this should move the color from cache to data
        m.visual.vertex_colors[0] = test_color_2
        assert (m.visual.vertex_colors[0] == test_color_2).all()
        assert (m.visual.vertex_colors[1] != test_color_2).any()
        assert m.visual.kind == "vertex"
        assert m.visual.defined
        assert not m.visual.transparency

        m.visual.vertex_colors[1] = test_color_transparent
        assert m.visual.transparency

        test = (g.random((len(m.faces), 4)) * 255).astype(g.np.uint8)
        m.visual.face_colors = test
        assert bool((m.visual.face_colors == test).all())
        assert m.visual.kind == "face"

        test = (g.random((len(m.vertices), 4)) * 255).astype(g.np.uint8)
        m.visual.vertex_colors = test
        assert bool((m.visual.vertex_colors == test).all())
        assert m.visual.kind == "vertex"

        test = (g.random(4) * 255).astype(g.np.uint8)
        m.visual.face_colors = test
        assert bool((m.visual.vertex_colors == test).all())
        assert m.visual.kind == "face"
        m.visual.vertex_colors[0] = [0, 0, 0, 0]
        assert m.visual.kind == "vertex"

        test = (g.random(4) * 255).astype(g.np.uint8)
        m.visual.vertex_colors = test
        assert bool((m.visual.face_colors == test).all())
        assert m.visual.kind == "vertex"
        m.visual.face_colors[:2] = (g.random((2, 4)) * 255).astype(g.np.uint8)
        assert m.visual.kind == "face"

    def test_smooth(self):
        """
        Make sure cached smooth model is dumped if colors are changed
        """
        m = g.get_mesh("featuretype.STL")

        # will put smoothed mesh into visuals cache
        s = m.smooth_shaded
        # every color should be default color
        assert g.np.ptp(s.visual.face_colors, axis=0).max() == 0

        # set one face to a different color
        m.visual.face_colors[0] = [255, 0, 0, 255]

        # cache should be dumped yo
        s1 = m.smooth_shaded
        assert g.np.ptp(s1.visual.face_colors, axis=0).max() != 0

        # do the same check on vertex color
        m = g.get_mesh("featuretype.STL")
        s = m.smooth_shaded
        # every color should be default color
        assert g.np.ptp(s.visual.vertex_colors, axis=0).max() == 0
        m.visual.vertex_colors[g.np.arange(10)] = [255, 0, 0, 255]
        s1 = m.smooth_shaded
        assert g.np.ptp(s1.visual.face_colors, axis=0).max() != 0

    def test_vertex(self):
        m = g.get_mesh("torus.STL")

        m.visual.vertex_colors = [100, 100, 100, 255]

        assert len(m.visual.vertex_colors) == len(m.vertices)

    def test_conversion(self):
        m = g.get_mesh("machinist.XAML")
        assert m.visual.kind == "face"

        # unmerge vertices so we don't get average colors
        m.unmerge_vertices()

        # store initial face colors
        initial = g.deepcopy(m.visual.face_colors.copy())

        # assign averaged vertex colors as default
        m.visual.vertex_colors = m.visual.vertex_colors
        assert m.visual.kind == "vertex"

        m.visual._cache.clear()
        assert g.np.allclose(initial, m.visual.face_colors)

    def test_interpolate(self):
        """
        Check our color interpolation
        """
        values = g.np.array([-1.0, 0.0, 1.0, 2.0])
        # should clamp
        colors = g.trimesh.visual.linear_color_map(values)

        assert g.np.allclose(colors[0], [255, 0, 0, 255])
        assert g.np.allclose(colors[1], [255, 0, 0, 255])
        assert g.np.allclose(colors[2], [0, 255, 0, 255])
        assert g.np.allclose(colors[3], [0, 255, 0, 255])

        # should scale to range
        colors = g.trimesh.visual.interpolate(values)
        assert g.np.allclose(colors[0], [255, 0, 0, 255])
        # scaled to range not clamped
        assert not g.np.allclose(colors[1], [255, 0, 0, 255])
        assert not g.np.allclose(colors[2], [0, 255, 0, 255])
        # end of range
        assert g.np.allclose(colors[3], [0, 255, 0, 255])

        # try interpolating with matplotlib color maps
        try:
            colors = g.trimesh.visual.interpolate(values, "viridis")
        except ImportError:
            # if matplotlib isn't installed
            return
        # check shape and type for matplotlib cmaps
        assert colors.shape == (len(values), 4)
        assert colors.dtype == g.np.uint8
        # every color should differ
        assert (colors[:-1] != colors[1:]).any(axis=1).all()

    def test_uv_to_color(self):
        try:
            import PIL.Image
        except ImportError:
            return

        # n_vertices = 100
        uv = g.np.array([[0.25, 0.2], [0.4, 0.5]], dtype=float)
        texture = g.np.arange(96, dtype=g.np.uint8).reshape(8, 4, 3)
        colors = g.trimesh.visual.uv_to_color(uv, PIL.Image.fromarray(texture))
        colors_expected = [[75, 76, 77, 255], [51, 52, 53, 255]]

        g.np.testing.assert_allclose(colors, colors_expected, rtol=0, atol=0)

    def test_uv_to_interpolated_color(self):
        try:
            import PIL.Image
        except ImportError:
            return

        uv = g.np.array([[0.25, 0.2], [0.4, 0.5]], dtype=float)

        tmp = g.np.arange(32)
        x, y = g.np.meshgrid(tmp, tmp)
        z = x + 128

        texture = g.np.stack([x, y, z], axis=-1).astype(g.np.uint8)

        img = PIL.Image.fromarray(texture)
        colors = g.trimesh.visual.uv_to_interpolated_color(uv, img)

        # exact interpolated values before being converted to uint8
        colors_expected = [[7.75, 24.8, 128 + 7.75, 255], [12.4, 15.5, 128 + 12.4, 255]]

        assert g.np.allclose(colors, colors_expected, rtol=0, atol=1)

    def test_iterset(self):
        m = g.trimesh.creation.box()
        color = [100, 0, 0, 200]

        # facets should include every face
        assert set(g.np.hstack(m.facets)) == set(range(len(m.faces)))
        assert len(m.facets) * 2 == len(m.faces)

        for f in m.facets:
            m.visual.face_colors[f] = color

        assert g.np.allclose(m.visual.face_colors, color)

    def test_copy(self):
        s = g.trimesh.creation.uv_sphere().scene()
        s.geometry["geometry_0"].visual.face_colors[:, :3] = [0, 255, 0]

        a = s.geometry["geometry_0"]
        assert id(a) == id(s.geometry["geometry_0"])

        b = s.geometry["geometry_0"].copy()
        assert id(a) != id(b)

        assert g.np.allclose(a.visual.face_colors, b.visual.face_colors)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()
